"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only


FlowSequencer knows about the DE1 and imports it
DE1 needs to know about the FlowSequencer to be able to set stop-at targets.

Resolve by defining a restricted interface
"""
import enum
import multiprocessing
import time
from typing import Optional, Union

import pyDE1
from pyDE1.de1.c_api import API_MachineStates, MAX_FRAMES
from pyDE1.event_manager.payloads import EventPayload
from pyDE1.exceptions import (
    DE1APITypeError, DE1APIValueError, DE1APINotManagedHereException
)


logger = pyDE1.getLogger('FlowSequencer')

ModeControl = Union[
    'I_EspressoControl',
    'I_HotWaterControl',
    'I_HotWaterRinseControl',
    'I_SteamControl'
]


class FlowSequencer():
    """
    A singleton base class that helps break circular includes
    Returns the FlowSequencerImpl singleton
    """

    def __new__(cls, *args, **kwds):
        # By importing FlowSequencerImpl at run time,
        # potential import issues should be mitigated.
        import pyDE1.flow_sequencer_impl
        return pyDE1.flow_sequencer_impl.FlowSequencerImpl()

    database_queue: Optional[multiprocessing.Queue] = None
    
    @property
    def stop_at_weight_adjust(self):
        raise NotImplementedError

    @property
    def de1(self):
        raise NotImplementedError

    async def set_up_subscribers(self):
        raise NotImplementedError

    @property
    def scale_processor(self):
        raise NotImplementedError

    @property
    def active_state(self):
        raise NotImplementedError

    def active_control_for_state(self,
                                 state: API_MachineStates) -> ModeControl:
        raise NotImplementedError

    @property
    def active_control(self) -> ModeControl:
        raise NotImplementedError

    @property
    def sequence_start_time(self):
        raise NotImplementedError

    # async def stop_at_notify(self, stop_at: 'StopAtType',
    #                          action: 'StopAtNotificationAction',
    #                          target_value: Optional[float],
    #                          current_value: Optional[float],
    #                          active_state: API_MachineStates,
    #                          current_frame: Optional[int]):
    #     raise NotImplementedError

    def stop_at_time_set(self, state: API_MachineStates, duration: float):
        raise NotImplementedError

    def stop_at_volume_set(self, state: API_MachineStates, volume: float):
        raise NotImplementedError

    def stop_at_weight_set(self, state: API_MachineStates, weight: float):
        raise NotImplementedError

    def profile_can_override_stop_limits(self, state: API_MachineStates):
        raise NotImplementedError

    def profile_can_override_tank_temperature(self, state: API_MachineStates):
        raise NotImplementedError

    async def on_de1_nearly_ready(self) -> None:
        raise NotImplementedError


LAST_DROPS_MINIMUM_TIME_DEFAULT = 3.0  # seconds
FIRST_DROPS_THRESHOLD_DEFAULT = 0.0  # bar


class StopAtNotificationAction (enum.Enum):
    ENABLED = 'enabled'
    TRIGGERED = 'triggered'
    DISABLED = 'disabled'
    DE1CONTROLLED = 'de1 controlled'


class StopAtType (enum.Enum):
    TIME = 'time'
    VOLUME = 'volume'
    WEIGHT = 'weight'
    MOW  = 'move on by weight'


class StopAtNotification (EventPayload):
    """
    Enable, disable, trigger of the various stop-at conditions
    current_value is generally only set for StopAtNotificationAction.TRIGGERED

    ENABLED notifications are given even if the target is None as the target
    can be changed during the shot, at least when managed by the FlowSequencer
    (On-the-fly profile changes are not supported at this time.
     On-the-fly changes to steam duration have not been tested at this time.)
    """
    def __init__(self, stop_at: StopAtType,
                 action: StopAtNotificationAction,
                 target_value: Optional[float] = None,
                 current_value: Optional[float] = None,
                 active_state: API_MachineStates = API_MachineStates.NoRequest,
                 current_frame: Optional[int] = None):
        now = time.time()
        super(StopAtNotification, self).__init__(
            arrival_time=now,
            create_time=now
        )
        self._version = "1.1.0"
        self.stop_at = stop_at
        self.action = action
        self.target_value = target_value
        self.current_value = current_value
        self.active_state = active_state
        self.current_frame = current_frame


class AutoTareNotificationAction (enum.Enum):
    ENABLED = 'enabled'
    DISABLED = 'disabled'


class AutoTareNotification (EventPayload):

    def __init__(self, action: AutoTareNotificationAction):
        now = time.time()
        super(AutoTareNotification, self).__init__(
            arrival_time=now,
            create_time=now
        )
        self._version = "1.0.0"
        self.action = action


class BaseModeControl:
    """
    Generic holder for parameters common to all four ModeControl objects,
    Espresso, HotWater, HotWaterFlush, and Steam
    """

    def __init__(self, disable_auto_tare: bool = False):
        self._disable_auto_tare = None
        self.disable_auto_tare = disable_auto_tare

    @property
    def disable_auto_tare(self):
        return self._disable_auto_tare

    @disable_auto_tare.setter
    def disable_auto_tare(self, value):
        if not isinstance(value, bool):
            raise DE1APITypeError(
                f"disable_auto_tare must be a bool, not {value}"
            )
        self._disable_auto_tare = value


def validate_stop_at(value):
    if value is not None:
        if value == 0:
            value = None
            logger.info(
                "Deprecated use of 0 for stop-at or move-on, "
                "replaced by None")
        elif value < 0:
            raise DE1APIValueError(
                "Stop-at and move-on values need to be non-negative, "
                f"not '{value}'")
    return value


class StopAtTimeControl:

    def __init__(self, stop_at_time: Optional[float] = None):
        # Mix-in, call super from concrete instance
        self._stop_at_time = None
        try:
            self.stop_at_time = stop_at_time
        except DE1APINotManagedHereException:
            self._stop_at_time = stop_at_time

    @property
    def stop_at_time(self):
        return self._stop_at_time

    @stop_at_time.setter
    def stop_at_time(self, value):
        self._stop_at_time = validate_stop_at(value)


class StopAtVolumeControl:

    def __init__(self, stop_at_volume: Optional[float] = None):
        # Mix-in, call super from concrete instance
        self._stop_at_volume = None
        self.stop_at_volume = stop_at_volume

    @property
    def stop_at_volume(self):
        return self._stop_at_volume

    @stop_at_volume.setter
    def stop_at_volume(self, value):
        self._stop_at_volume = validate_stop_at(value)


class StopAtWeightControl:

    def __init__(self, stop_at_weight: Optional[float] = None):
        # Mix-in, call super from concrete instance
        self._stop_at_weight = None
        self.stop_at_weight = stop_at_weight

    @property
    def stop_at_weight(self):
        return self._stop_at_weight

    @stop_at_weight.setter
    def stop_at_weight(self, value):
        self._stop_at_weight = validate_stop_at(value)


class MoveOnWeightControl:

    def __init__(self, mow_all_frames: Optional[list[Optional[float]]] = None):
        # Mix-in, call super from concrete instance
        self._mow_all_frames = []
        self.mow_all_frames = mow_all_frames

    @property
    def mow_all_frames(self) -> list[Optional[Union[int, float]]]:
        return self._mow_all_frames.copy()

    @mow_all_frames.setter
    def mow_all_frames(self,
                       value_list: Optional[list[Optional[Union[int, float]]]]):
        if value_list is None:
            value_list = []
        if not isinstance(value_list, list):
            raise DE1APIValueError(
                f"Move-on-weight setter expecting a list, not '{value_list}'")
        if len(value_list):
            map(validate_stop_at, value_list)
        self._mow_all_frames = value_list

    def mow_get_frame(self, frame_number: int) -> Optional[Union[int, float]]:
        retval = None
        if frame_number is not None:
            try:
                retval = self._mow_all_frames[frame_number]
            except IndexError:
                pass
        return retval

    def mow_set_frame(self, frame_number: int,
                      value: Optional[Union[int, float]]):
        if frame_number > MAX_FRAMES:
            raise DE1APIValueError(
                f"Request to set mow for frame {frame_number} > MAX_FRAMES")
        if frame_number is None:
            raise DE1APIValueError(
                f"Request to set mow for frame 'None'")
        if frame_number < 0:
            raise DE1APIValueError(
                f"Request to set mow for frame {frame_number} < 0")
        validate_stop_at(value)
        while len(self._mow_all_frames) < frame_number:
            self._mow_all_frames.append(None)
        self._mow_all_frames[frame_number] = value


class ProfileOverrideControl:

    def __init__(self,
                 profile_can_override_stop_limits: bool = True,
                 profile_can_override_tank_temperature: bool = True,
                 ):
        self._profile_can_override_stop_limits \
            = profile_can_override_stop_limits
        self._profile_can_override_tank_temperature \
            = profile_can_override_tank_temperature

    @property
    def profile_can_override_stop_limits(self):
        return self._profile_can_override_stop_limits

    @profile_can_override_stop_limits.setter
    def profile_can_override_stop_limits(self, value):
        if not isinstance(value, bool):
            raise DE1APITypeError(
                "profile_can_override_stop_limits must be a bool, "
                f"not {value}"
            )
        self._profile_can_override_stop_limits = value

    @property
    def profile_can_override_tank_temperature(self):
        return self._profile_can_override_tank_temperature

    @profile_can_override_tank_temperature.setter
    def profile_can_override_tank_temperature(self, value):
        if not isinstance(value, bool):
            raise DE1APITypeError(
                "profile_can_override_tank_temperature must be a bool, "
                f"not {value}"
            )
        self._profile_can_override_tank_temperature = value


class EspressoDropsControl:

    def __init__(self,
                 first_drops_threshold: Optional[float] = \
                         FIRST_DROPS_THRESHOLD_DEFAULT,
                 last_drops_minimum_time: float = \
                         LAST_DROPS_MINIMUM_TIME_DEFAULT,
                 ):
        self.first_drops_threshold = first_drops_threshold
        self.last_drops_minimum_time = last_drops_minimum_time

    @property
    def first_drops_threshold(self):
        return self._first_drops_threshold

    @first_drops_threshold.setter
    def first_drops_threshold(self, value):
        if value and not (0 <= value <= 10):
            raise DE1APIValueError(
                f"first_drops_threshold not 0 <= {value} <= 10")
        self._first_drops_threshold = value

    @property
    def last_drops_minimum_time(self):
        return self._last_drops_minimum_time

    @last_drops_minimum_time.setter
    def last_drops_minimum_time(self, value):
        if value < 0:
            raise DE1APIValueError(
                f"last_drops_minimum_time less than zero ({value}")
        self._last_drops_minimum_time = value


# These are primarily defined here as interfaces and some "real" implementation
# details in flow_sequencer_impl.py to break circular dependencies
# between the DE1 and FlowSequencer definitions.

class I_EspressoControl (BaseModeControl,
                         StopAtTimeControl,
                         StopAtVolumeControl,
                         StopAtWeightControl,
                         MoveOnWeightControl,
                         EspressoDropsControl,
                         ProfileOverrideControl):

    def __init__(self, disable_auto_tare: bool = False,
                 stop_at_time: Optional[float] = None,
                 stop_at_volume: Optional[float] = None,
                 stop_at_weight: Optional[float] = None,
                 profile_can_override_stop_limits: bool = True,
                 profile_can_override_tank_temperature: bool = True,
                 first_drops_threshold: Optional[float] = \
                         FIRST_DROPS_THRESHOLD_DEFAULT,
                 last_drops_minimum_time: float = \
                         LAST_DROPS_MINIMUM_TIME_DEFAULT,
                 ):
        BaseModeControl.__init__(self, disable_auto_tare=disable_auto_tare)
        StopAtTimeControl.__init__(self, stop_at_time=stop_at_time)
        StopAtVolumeControl.__init__(self, stop_at_volume=stop_at_volume)
        StopAtWeightControl.__init__(self, stop_at_weight=stop_at_weight)
        # As the control is usually instantiated before a DE1 is connected
        # and before a profile is uploaded, the default here is OK
        MoveOnWeightControl.__init__(self, mow_all_frames=None)
        EspressoDropsControl.__init__(self,
                            first_drops_threshold=first_drops_threshold,
                            last_drops_minimum_time=last_drops_minimum_time)
        ProfileOverrideControl.__init__(self,
                            profile_can_override_stop_limits= \
                                profile_can_override_stop_limits,
                            profile_can_override_tank_temperature= \
                                profile_can_override_tank_temperature)


class I_HotWaterControl (BaseModeControl,
                         StopAtTimeControl,
                         StopAtVolumeControl,
                         StopAtWeightControl):

    def __init__(self, disable_auto_tare: bool = False,
                 stop_at_time: Optional[float] = None,
                 stop_at_volume: Optional[float] = None,
                 stop_at_weight: Optional[float] = None,
                 ):
        BaseModeControl.__init__(self, disable_auto_tare=disable_auto_tare)
        StopAtTimeControl.__init__(self, stop_at_time=stop_at_time)
        StopAtVolumeControl.__init__(self, stop_at_volume=stop_at_volume)
        StopAtWeightControl.__init__(self, stop_at_weight=stop_at_weight)


class I_HotWaterRinseControl (BaseModeControl,
                              StopAtTimeControl,
                              ):

    def __init__(self, disable_auto_tare: bool = False,
                 stop_at_time: Optional[float] = None,
                 ):
        BaseModeControl.__init__(self, disable_auto_tare=disable_auto_tare)
        StopAtTimeControl.__init__(self, stop_at_time=stop_at_time)

    # stop_at_time overridden by HotWaterRinseControl


class I_SteamControl (BaseModeControl,
                      StopAtTimeControl,
                      ):

    def __init__(self, disable_auto_tare: bool = False,
                 stop_at_time: Optional[float] = None,
                 ):
        BaseModeControl.__init__(self, disable_auto_tare=disable_auto_tare)
        StopAtTimeControl.__init__(self, stop_at_time=stop_at_time)

    # stop_at_time overridden by SteamControl



"""
Copyright © 2021-2022 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

Provide a mapping from internal objects and properties, sufficient for both
creation of outgoing dict, as well as validation of incoming requests
(Resource, key, target object, property name, value type, min, max, post_execute)

As of MAPPING_VERSION 4.0.0 direct use of the class or a stub has been replaced
by use of an enum. Consumers will need to map the enum to the class.
This is due to challenges in determining if the class or the stub is needed.
"""

import importlib.util
import inspect
import sys
from enum import Enum, auto
from typing import Optional, Union, NamedTuple

import pyDE1
from pyDE1.de1.c_api import (
    PackedAttr, MMR0x80LowAddr,
    SetTime, ShotSettings, Versions, WaterLevels,
)
from pyDE1.dispatcher.resource import (
    Resource, RESOURCE_VERSION, DE1ModeEnum, ConnectivityEnum
)

MAPPING_VERSION = "5.0.0"

logger = pyDE1.getLogger('Inbound.Mapping')

# There are a handful of requests related to the DE1 and Scale
# that can be or must be handled even if the device is not ready.
# At this time, they include those in DE1_ID and SCALE_ID
# Rather than introduce great complexity to accommodate those exceptions,
#   * Move the remaining, related Scale access to ScaleProcessor
#   * Ensure that the DE1 and ScaleProcessor return meaningful values
#   * Ensure that if None is returned, that the IsAt indicates Optional[]
#   * Add the "if_not_ready" element to IsAt and check in requires_*


class TO (Enum):
    DE1 = auto()
    DiscoveredDevices = auto()
    FlowSequencer = auto()
    Scale = auto()
    ScaleProcessor = auto()
    

class IsAt(NamedTuple):
    # TODO: This list appears to miss None for Resource.SCAN (scanner.py)
    #       Maybe reference by the module?
    target: Union[type(TO),
                  type(PackedAttr),
                  type(MMR0x80LowAddr),]
    attr_path: Optional[str]  # '' for MMR0x80LowAddr, None for write_only,
    v_type: type  # expected value type for type-checking
    setter_path: Optional[str] = None   # If not a property and a different path
                                        # path of setter relative to target
    read_only: Optional[bool] = False
    # write_only: attr_path=None and setter_path=setter_attribute_name
    # internal_type used for conversion to enum
    internal_type: Optional[Union[type(DE1ModeEnum),
                                  type(ConnectivityEnum)]] = None
    # A handful of things, such as connectivity, don't need "ready"
    if_not_ready: bool = False

    @property
    def requires_connected_de1(self) -> bool:
        retval = (
            (self.target == TO.DE1 and not self.if_not_ready)
            or isinstance(self.target, MMR0x80LowAddr)
            or (inspect.isclass(self.target)
                and issubclass(self.target, PackedAttr))
        )
        return retval

    @property
    def requires_connected_scale(self) -> bool:
        retval = (
            (self.target == TO.Scale and not self.if_not_ready)
        )
        return retval


def mapping_requires(mapping: dict) -> dict:
    results = {
        'DE1': False,
        'Scale': False
    }
    return _mapping_requires_inner(mapping, results)


def _mapping_requires_inner(mapping: dict, results: dict) -> dict:
    for val in mapping.values():
        if isinstance(val, IsAt):
            if val.requires_connected_de1:
                results['DE1'] = True
            if val.requires_connected_scale:
                results['Scale'] = True

        if isinstance(val, dict):
            _mapping_requires_inner(val, results)

        # TODO: Maybe one day generify this
        if results['DE1'] and results['Scale']:
            break

    return results


# Helper to populate an IsAt for a PackedAttr
def from_packed_attr(packed_attr: PackedAttr, attr_path: str, v_type: type,
                     setter_path: Optional[str] = None):
    # get_cuuid will raise if not over-the-wire
    cuuid = get_cuuid(packed_attr)
    can_read = cuuid.can_read or cuuid.can_notify
    can_write = cuuid.can_write
    if attr_path is not None and not can_read:
        raise DE1APIValueError(
            f"Un-readable PackedAttr {packed_attr} "
            f"with attr_path set to '{attr_path}'"
        )

    return IsAt(
        target=packed_attr,
        attr_path=attr_path,
        v_type=v_type,
        setter_path=setter_path,
        read_only=not can_read,
    )


# Helper to populate an IsAt for MMR
def from_mmr(mmr: MMR0x80LowAddr, v_type: type):
    # This doesn't handle can't read and can't write
    return IsAt(
        target=mmr,
        attr_path='',
        v_type=v_type,
        read_only=not mmr.can_read
    )


MODULES_FOR_VERSIONS = (
    'pyDE1',
    'aiosqlite',
    'bleak',
    'paho-mqtt',
    'PyYAML',
    'requests',
    # Retained for backward compatibility - consider as deprecated in this list
    'asyncio-mqtt',
)

import importlib.metadata  # Used for module-version lookup only


def module_versions():
    retval = {}
    for module in MODULES_FOR_VERSIONS:
        try:
            retval[module] = importlib.metadata.version(module)
        except importlib.metadata.PackageNotFoundError:
            retval[module] = None
    return retval


MAPPING = {}

MAPPING[Resource.VERSION] = {
    'resource_version': RESOURCE_VERSION,
    'mapping_version': MAPPING_VERSION,
    'platform': sys.platform,
    'python': sys.version,
    'python_info': {
        'major': sys.version_info.major,
        'minor': sys.version_info.minor,
        'micro': sys.version_info.micro,
        'releaselevel': sys.version_info.releaselevel,
        'serial': sys.version_info.serial,
    },
    'module_versions': module_versions()
}

# "Specials" -- content-only

MAPPING[Resource.DE1_PROFILE] = IsAt(target=TO.DE1, attr_path=None,
                                     setter_path='upload_json_v2_profile',
                                     v_type=Union[bytes, bytearray])

MAPPING[Resource.DE1_PROFILE_STORE] = IsAt(target=TO.DE1, attr_path=None,
                                     setter_path='store_json_v2_profile',
                                     v_type=Union[bytes, bytearray])

MAPPING[Resource.DE1_FIRMWARE] = IsAt(target=TO.DE1, attr_path=None,
                                      setter_path='upload_firmware_from_content',
                                      v_type=Union[bytes, bytearray])

MAPPING[Resource.DE1_FIRMWARE_CANCEL] = IsAt(target=TO.DE1, attr_path=None,
                                             setter_path='cancel_firmware_api',
                                             v_type=Union[bytes, bytearray])

# TODO: How to reference a module-level subroutine? Use module as target?

MAPPING[Resource.SCAN] = {
    # Passing a number sets the timeout. Passing null takes the default.
    'begin': IsAt(target=None, attr_path=None,
                  setter_path='scan_from_api',
                  v_type=Optional[Union[int, float]])
}

# Note: This is an async getter because of the lock
MAPPING[Resource.SCAN_DEVICES] = {
    'devices': IsAt(target=TO.DiscoveredDevices, attr_path='devices_for_json',
                    read_only=True,
                    v_type=dict)
}

# Work from leaves back, so can be "included" by reference

# For now, name is not writable
MAPPING[Resource.DE1_ID] = {
    'name': IsAt(target=TO.DE1, attr_path='name', v_type=Optional[str],
                 read_only=True,
                 if_not_ready=True),
    # The ID value 'scan' will perform a scan and capture the first found
    'id': IsAt(target=TO.DE1, attr_path='address', v_type=Optional[str],
               setter_path='change_de1_to_id',
               if_not_ready=True),
}

# NB: A single-entry tuple needs to end with a comma

MAPPING[Resource.DE1_MODE] = {
    'mode': IsAt(target=TO.DE1, attr_path=None, setter_path='mode_setter',
                 v_type=str, internal_type=DE1ModeEnum),
}

MAPPING[Resource.DE1_STATE] = {
    'state': IsAt(target=TO.DE1, attr_path='state_getter',
                  read_only=True,
                  v_type=str, internal_type=str),
}

MAPPING[Resource.DE1_FEATURE_FLAGS] = {
    'feature_flags': IsAt(target=TO.DE1, attr_path='feature_flags',
                          v_type=str, read_only=True),
}

# DE1_PROFILE = 'de1/profile'
# DE1_PROFILES = 'de1/profiles'

MAPPING[Resource.DE1_PROFILE_ID] = {
    'id': IsAt(target=TO.DE1, attr_path='profile_id', v_type=str,
               setter_path='set_profile_by_id')
}

# DE1_FIRMWARE = 'de1/firmware'
# DE1_FIRMWARES = 'de1/firmwares'

MAPPING[Resource.DE1_FIRMWARE_ID] = {
    'id': IsAt(target=MMR0x80LowAddr.CPU_FIRMWARE_BUILD,
               attr_path='', v_type=int),
}

MAPPING[Resource.DE1_CONNECTIVITY] = {
    'mode': IsAt(target=TO.DE1, attr_path='connectivity',
                 setter_path='connectivity_setter', v_type=str,
                 internal_type=ConnectivityEnum,
                 if_not_ready=True),
}

# DE1_CONTROL = 'de1/control' -- aggregate
# TODO: Work through how to get the aggregate to operate


MAPPING[Resource.DE1_CONTROL_ESPRESSO] = {
    'stop_at_time': IsAt(target=TO.FlowSequencer,
                         attr_path='espresso_control.stop_at_time',
                         v_type=Optional[float]),
    'stop_at_volume': IsAt(target=TO.FlowSequencer,
                           attr_path='espresso_control.stop_at_volume',
                           v_type=Optional[float]),
    'stop_at_weight': IsAt(target=TO.FlowSequencer,
                           attr_path='espresso_control.stop_at_weight',
                           v_type=Optional[float]),
    'move_on_weight': IsAt(target=TO.FlowSequencer,
                           attr_path='espresso_control.mow_all_frames',
                           v_type=list[Optional[Union[int, float]]]),
    'disable_auto_tare': IsAt(target=TO.FlowSequencer,
                              attr_path='espresso_control.disable_auto_tare',
                              v_type=bool),

    'profile_can_override_stop_limits':
        IsAt(target=TO.FlowSequencer,
             attr_path='espresso_control.profile_can_override_stop_limits',
             v_type=bool),
    'profile_can_override_tank_temperature':
        IsAt(target=TO.FlowSequencer,
             attr_path='espresso_control.profile_can_override_tank_temperature',
             v_type=bool),
    'first_drops_threshold':
        IsAt(target=TO.FlowSequencer,
             attr_path='espresso_control.first_drops_threshold',
             v_type=Optional[float]),
    'last_drops_minimum_time':
        IsAt(target=TO.FlowSequencer,
             attr_path='espresso_control.last_drops_minimum_time',
             v_type=float),
}

MAPPING[Resource.DE1_CONTROL_STEAM] = {
    'stop_at_time': IsAt(target=ShotSettings, attr_path='TargetSteamLength', v_type=int),
    'stop_at_volume': IsAt(target=TO.FlowSequencer,
                           attr_path='steam_control.stop_at_volume',
                           v_type=Optional[float]),
    'stop_at_weight': IsAt(target=TO.FlowSequencer,
                           attr_path='steam_control.stop_at_weight',
                           v_type=Optional[float]),
    'disable_auto_tare': IsAt(target=TO.FlowSequencer,
                              attr_path='steam_control.disable_auto_tare',
                              v_type=bool),
}

MAPPING[Resource.DE1_CONTROL_HOT_WATER] = {
    'stop_at_time': IsAt(target=ShotSettings,
                         attr_path='TargetHotWaterLength',
                         v_type=int),
    'stop_at_volume': IsAt(target=ShotSettings,
                           attr_path='TargetHotWaterVol',
                           v_type=int),
    'stop_at_weight': IsAt(target=TO.FlowSequencer,
                           attr_path='hot_water_control.stop_at_weight',
                           v_type=Optional[float]),
    'disable_auto_tare': IsAt(target=TO.FlowSequencer,
                              attr_path='hot_water_control.disable_auto_tare',
                              v_type=bool),
    'temperature': IsAt(target=ShotSettings,
                        attr_path='TargetHotWaterTemp', v_type=int),
}

MAPPING[Resource.DE1_CONTROL_HOT_WATER_RINSE] = {
    'stop_at_time':
        IsAt(target=TO.FlowSequencer,
             attr_path='hot_water_rinse_control.stop_at_time',
             setter_path='hot_water_rinse_control.stop_at_time_set_async',
             v_type=Optional[float]),
    'stop_at_volume':
        IsAt(target=TO.FlowSequencer,
             attr_path='hot_water_rinse_control.stop_at_volume',
             v_type=Optional[float]),
    'stop_at_weight':
        IsAt(target=TO.FlowSequencer,
             attr_path='hot_water_rinse_control.stop_at_weight',
             v_type=Optional[float]),
    'disable_auto_tare':
        IsAt(target=TO.FlowSequencer,
             attr_path='hot_water_rinse_control.disable_auto_tare', v_type=bool),
    'temperature':
        IsAt(target=MMR0x80LowAddr.FLUSH_TEMP, attr_path='', v_type=float),
    'flow':
        IsAt(target=MMR0x80LowAddr.FLUSH_FLOW_RATE, attr_path='', v_type=float),
}

MAPPING[Resource.DE1_CONTROL_TANK_WATER_THRESHOLD] = {
    'temperature': IsAt(target=MMR0x80LowAddr.TANK_TEMP, attr_path='', v_type=int),
}

# DE1_SETTING = 'de1/setting' -- aggregate

MAPPING[Resource.DE1_SETTING_AUTO_OFF_TIME] = {
    'time': IsAt(target=TO.DE1, attr_path='auto_off_time',
                 v_type=Optional[float]),
}

MAPPING[Resource.DE1_SETTING_FAN_THRESHOLD] = {
    'temperature': IsAt(target=MMR0x80LowAddr.FAN_THRESHOLD, attr_path='', v_type=int),
}

MAPPING[Resource.DE1_SETTING_START_FILL_LEVEL] = {
    'start_fill_level': IsAt(target=WaterLevels, attr_path='StartFillLevel', v_type=int),
}

MAPPING[Resource.DE1_SETTING_BEFORE_FLOW] = {
    'heater_phase1_flow': IsAt(target=MMR0x80LowAddr.HEATER_UP1_FLOW, attr_path='', v_type=float),
    'heater_phase2_flow': IsAt(target=MMR0x80LowAddr.HEATER_UP2_FLOW, attr_path='', v_type=float),
    'heater_phase2_timeout': IsAt(target=MMR0x80LowAddr.HEATER_UP2_TIMEOUT, attr_path='', v_type=float),
    'heater_idle_temperature': IsAt(target=MMR0x80LowAddr.WATER_HEATER_IDLE_TEMP, attr_path='', v_type=float),
}

MAPPING[Resource.DE1_SETTING_STEAM] = {
    'temperature': IsAt(target=ShotSettings, attr_path='TargetSteamTemp', v_type=int),
    'flow': IsAt(target=MMR0x80LowAddr.TARGET_STEAM_FLOW, attr_path='', v_type=float),
    'high_flow_time': IsAt(target=MMR0x80LowAddr.STEAM_START_SECS, attr_path='', v_type=float),
    'purge_deferred': IsAt(target=MMR0x80LowAddr.STEAM_PURGE_MODE, attr_path='', v_type=bool),
}

# TODO: What is ShotSettings.TargetGroupTemp in the current DE1 firmware
#       and where does it really belong?
MAPPING[Resource.DE1_SETTING_TARGET_GROUP_TEMP] = {
    'temperature': IsAt(target=ShotSettings, attr_path='TargetGroupTemp', v_type=int),
}

MAPPING[Resource.DE1_SETTING_USB_OUTLET] = {
    'enabled':  IsAt(target=MMR0x80LowAddr.ALLOW_USB_CHARGING,
                     attr_path='', v_type=bool),
}

MAPPING[Resource.DE1_SETTING_REFILL_KIT] = {
    'present':  IsAt(target=MMR0x80LowAddr.REFILL_KIT_PRESENT,
                     attr_path='', v_type=bool),
}

# NB: Not exposed at this time
MAPPING[Resource.DE1_DEPRECATED] = {
    'old_espresso_vol': IsAt(target=ShotSettings,
                             attr_path='TargetEspressoVol', v_type=int),
    'steam_fast_start': IsAt(target=ShotSettings,
                       attr_path='steam_setting_fast_start', v_type=bool),
    'steam_high_power': IsAt(target=ShotSettings,
                       attr_path='steam_setting_high_power', v_type=bool),
}

MAPPING[Resource.DE1_SETTING_TIME] = {
    'timestamp': IsAt(target=SetTime, attr_path='Timestamp', v_type=int),
}

# None exposed yet
MAPPING[Resource.DE1_PARAMETER_SET] = {
}

MAPPING[Resource.DE1_READ_ONCE] = {
    'hw_config_hexstr': IsAt(target=MMR0x80LowAddr.HW_CONFIG, attr_path='', v_type=str),
    'model_hexstr': IsAt(target=MMR0x80LowAddr.MODEL, attr_path='', v_type=str),
    'cpu_board_model': IsAt(target=MMR0x80LowAddr.CPU_BOARD_MODEL, attr_path='', v_type=float),
    'firmware_model': IsAt(target=MMR0x80LowAddr.V13_MODEL, attr_path='', v_type=str),
    'firmware_build_number': IsAt(target=MMR0x80LowAddr.CPU_FIRMWARE_BUILD, attr_path='', v_type=int),
    'ghc_info': IsAt(target=MMR0x80LowAddr.GHC_INFO, attr_path='', v_type=str),  # See MMRGHCInfoBitMask
    'serial_number_hexstr': IsAt(target=MMR0x80LowAddr.SERIAL_NUMBER, attr_path='', v_type=str),
    'heater_voltage': IsAt(target=MMR0x80LowAddr.HEATER_VOLTAGE, attr_path='', v_type=int),

    'version_ble': {
        'api': IsAt(target=Versions, attr_path='BLEVersion.APIVersion', v_type=int),
        'release': IsAt(target=Versions, attr_path='BLEVersion.Release', v_type=float),
        'commits': IsAt(target=Versions, attr_path='BLEVersion.Commits', v_type=int),
        'changes': IsAt(target=Versions, attr_path='BLEVersion.Changes', v_type=int),
        'blesha_hexstr': IsAt(target=Versions, attr_path='BLEVersion.BLESha', v_type=str),
    },

    'version_lv': {
        'api': IsAt(target=Versions, attr_path='LVVersion.APIVersion', v_type=int),
        'release': IsAt(target=Versions, attr_path='LVVersion.Release', v_type=float),
        'commits': IsAt(target=Versions, attr_path='LVVersion.Commits', v_type=int),
        'changes': IsAt(target=Versions, attr_path='LVVersion.Changes', v_type=int),
        'blesha_hexstr': IsAt(target=Versions, attr_path='LVVersion.BLESha', v_type=str),
    },
}

MAPPING[Resource.DE1_CALIBRATION_FLOW_MULTIPLIER] = {
    'multiplier': IsAt(target=MMR0x80LowAddr.CAL_FLOW_EST,
                       attr_path='', v_type=float),
}

MAPPING[Resource.DE1_CALIBRATION_LINE_FREQUENCY] = {
    'hz': IsAt(target=TO.DE1, attr_path='line_frequency', v_type=int)
}

MAPPING[Resource.DE1_PRESENCE] = {
    'user': IsAt(target=MMR0x80LowAddr.USER_PRESENT, attr_path='', v_type=bool)

}

MAPPING[Resource.SCALE_ID] = {
    'name': IsAt(target=TO.ScaleProcessor, attr_path='scale_name',
                 v_type=Optional[str],
                 read_only=True,
                 if_not_ready=True),
    # The ID value 'scan' will perform a scan and capture the first found
    'id': IsAt(target=TO.ScaleProcessor, attr_path='scale_address',
               v_type=Optional[str],
               setter_path='change_scale_to_id',
               if_not_ready=True),
    'type': IsAt(target=TO.ScaleProcessor, attr_path='scale_type',
                 v_type=Optional[str],
                 read_only=True,
                 if_not_ready=True),
}

MAPPING[Resource.SCALE_CONNECTIVITY] = {
    'mode': IsAt(target=TO.ScaleProcessor, attr_path='scale_connectivity',
                 setter_path="connectivity_setter", v_type=str,
                 internal_type=ConnectivityEnum,
                 if_not_ready=True),
}

MAPPING[Resource.SCALE_TARE] = {
    'tare': IsAt(target=TO.Scale, attr_path='', setter_path='tare_with_bool',
                 v_type=Optional[bool])  # Accommodate None as False
}

MAPPING[Resource.SCALE_DISPLAY] = {
    'display_on': IsAt(target=TO.Scale, attr_path='', setter_path='display_bool',
                       v_type=Optional[bool])  # Accommodate None as False
}

MAPPING[Resource.DE1_CONTROL] = {
    'espresso': MAPPING[Resource.DE1_CONTROL_ESPRESSO],
    'steam': MAPPING[Resource.DE1_CONTROL_STEAM],
    'hot_water': MAPPING[Resource.DE1_CONTROL_HOT_WATER],
    'hot_water_rinse': MAPPING[Resource.DE1_CONTROL_HOT_WATER_RINSE],
    'tank_water_threshold': MAPPING[Resource.DE1_CONTROL_TANK_WATER_THRESHOLD],
}

MAPPING[Resource.DE1_SETTING] = {
    'auto_off_time': MAPPING[Resource.DE1_SETTING_AUTO_OFF_TIME],
    'fan_threshold': MAPPING[Resource.DE1_SETTING_FAN_THRESHOLD],
    'start_fill_level': MAPPING[Resource.DE1_SETTING_START_FILL_LEVEL],
    'before_flow': MAPPING[Resource.DE1_SETTING_BEFORE_FLOW],
    'target_group_temp': MAPPING[Resource.DE1_SETTING_TARGET_GROUP_TEMP],
    'steam': MAPPING[Resource.DE1_SETTING_STEAM],
    'time': MAPPING[Resource.DE1_SETTING_TIME],
    'usb_outlet': MAPPING[Resource.DE1_SETTING_USB_OUTLET],
    'refill_kit': MAPPING[Resource.DE1_SETTING_REFILL_KIT],
}

MAPPING[Resource.DE1_CALIBRATION] = {
    'flow_multiplier': MAPPING[Resource.DE1_CALIBRATION_FLOW_MULTIPLIER],
    'line_frequency': MAPPING[Resource.DE1_CALIBRATION_LINE_FREQUENCY],
    # 'internal': Mapping[Resource.DE1_CALIBRATION_INTERNAL],
}

MAPPING[Resource.DE1] = {
    'id': MAPPING[Resource.DE1_ID],
    'mode': MAPPING[Resource.DE1_MODE],
    'state': MAPPING[Resource.DE1_STATE],
    # profile
    # profiles
    # firmware
    # firmwares
    'connectivity': MAPPING[Resource.DE1_CONNECTIVITY],
    'control': MAPPING[Resource.DE1_CONTROL],
    'setting': MAPPING[Resource.DE1_SETTING],
    'calibration': MAPPING[Resource.DE1_CALIBRATION],
    'parameter_set': MAPPING[Resource.DE1_PARAMETER_SET],
    'read_once': MAPPING[Resource.DE1_READ_ONCE],
}

MAPPING[Resource.SCALE] = {
    'id': MAPPING[Resource.SCALE_ID],
    'connectivity': MAPPING[Resource.SCALE_CONNECTIVITY],
    'tare': MAPPING[Resource.SCALE_TARE],
    'display': MAPPING[Resource.SCALE_DISPLAY],
}

MAPPING[Resource.FLOW_SEQUENCER_SETTING] = {}

MAPPING[Resource.FLOW_SEQUENCER_PARAMETER_SET] = {}

MAPPING[Resource.FLOW_SEQUENCER] = {
    'setting': MAPPING[Resource.FLOW_SEQUENCER_SETTING],
    'parameter_set': MAPPING[Resource.FLOW_SEQUENCER_PARAMETER_SET],
}



"""
Copyright © 2021, 2022 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

"Null" outbound processor that just keeps the queue clear,
counting the various messages by type,
logging every update_period seconds (and resetting the count)
"""

# Supervise:
#   nothing: loop.add_reader()

# Only import the minimal here, as it potentially ends up in all processes.

import enum
import multiprocessing
import multiprocessing.connection as mpc

import pyDE1.config


class OutboundMode (enum.Enum):
    EventPayload = enum.auto()
    LogRecord = enum.auto()


# See will_topic, below
class MQTTStatusText (enum.Enum):
    on_connection = 'Here'
    on_graceful_disconnect = 'Gone'
    on_will = 'Died'


def run_mqtt_outbound(master_config: pyDE1.config.Config,
                      log_queue: multiprocessing.Queue,
                      outbound_pipe: mpc.Connection,
                      mode: OutboundMode):

    pyDE1.config.config = master_config
    from pyDE1.config import config

    import asyncio
    import json
    import logging
    import os
    import time

    from typing import Callable
    from socket import gethostname

    import paho.mqtt.client as mqtt
    from paho.mqtt.client import MQTTv5, MQTT_CLEAN_START_FIRST_ONLY

    import pyDE1.pyde1_logging as pyde1_logging
    import pyDE1.shutdown_manager as sm

    from pyDE1.supervise import SupervisedTask

    if mode == OutboundMode.LogRecord:
        logger = pyDE1.getLogger('LogMQTT')
        will_subtopic = 'logging'
    else:
        logger = pyDE1.getLogger('Outbound')
        will_subtopic = 'notification'

    will_topic = "/".join((
        config.mqtt.TOPIC_ROOT, 'status/mqtt', will_subtopic,
    ))

    pyde1_logging.setup_queue_logging(config.logging, log_queue)
    pyde1_logging.config_logger_levels(config.logging)

    client_logger = logger.getChild('MQTTClient')
    client_logger.level = logging.INFO

    # https://github.com/eclipse/paho.mqtt.c/issues/864
    # Add support for Unix-domain sockets #864 (open issue)

    # MQTT_PROTOCOL_VERSION = asyncio_mqtt.client.ProtocolVersion.V5

    loop = asyncio.get_event_loop()
    loop.set_debug(True)

    def on_shutdown_underway_cleanup():
        logger.info("Watching for shutdown event")
        sm.shutdown_underway.wait()
        logger.info("Shutting down MQTT client")
        if mqtt_client.is_connected():
            mqtt_client.disconnect()
        mqtt_client.loop_stop()
        logger.info("Setting cleanup_complete")
        sm.cleanup_complete.set()

    on_shutdown_wait_task = loop.run_in_executor(
        None, on_shutdown_underway_cleanup)

    sm.attach_signal_handler_to_loop(sm.shutdown, loop)

    loop.set_exception_handler(sm.exception_handler)

    def _start_shutdown(sig = None):
        loop.create_task(sm.shutdown(sig, loop))

    async def heartbeat():
        hlog = pyDE1.getLogger(
            f"Heartbeat.{multiprocessing.current_process().name}")
        while not sm.shutdown_underway.is_set():
            await asyncio.sleep(10)
            hlog.debug("===== BEEP =====")

    heartbeat_task = SupervisedTask(heartbeat)

    def on_log_callback(client: mqtt.Client, userdata, level, buf):
        client_logger.info(f"CB: Log: level: {level} '{buf}' ({type(buf)})")

    def on_connect_callback(client, userdata, flags, reasonCode, properties):
        if reasonCode == 0:
            level = logging.INFO
        else:
            level = logging.ERROR
        client_logger.log(level,
            f"CB: Connect: flags: {flags}, reasonCode: {reasonCode}, "
            f"properties {properties}")
        # Split "action" from logging for clarity
        if reasonCode == 0:
            _send_on_connection_status()
        else:
            client_logger.critical(
                f"Connection to MQTT broker failed: {str(reasonCode)}, "
                "initiating process shutdown." )
            loop.call_soon_threadsafe(_start_shutdown)

    def _send_on_connection_status():
        mqtt_client.publish(
            topic=will_topic,
            payload=MQTTStatusText.on_connection.value,
            qos=0,
            retain=True,
        )

    def _send_on_graceful_disconnect():
        logger.info("Publishing graceful disconnect status")
        rc = mqtt_client.publish(
            topic=will_topic,
            payload=MQTTStatusText.on_graceful_disconnect.value,
            qos=0,
            retain=True,
        )
        rc.wait_for_publish()

    class ClientSendsGracefulDisconnect (mqtt.Client):

        def disconnect(self, *args, **kwargs):
            _send_on_graceful_disconnect()
            super(ClientSendsGracefulDisconnect,
                  self).disconnect(*args, **kwargs)

    def on_publish_callback(client, userdata, mid):
        client_logger.debug(f"CB: Published: mid: {mid}")

    # Caught exception in on_disconnect:
    #     on_disconnect_callback() missing 1 required positional argument:
    #         'properties'
    def on_disconnect_callback(client, userdata, reasonCode, properties=None):
        if sm.shutdown_underway.is_set():
            level = logging.INFO
        else:
            level = logging.ERROR
        client_logger.log(level, f"CB: Disconnect: reasonCode: {reasonCode}, "
                                 f"properties {properties}")

    def on_socket_open_callback(client, userdata, socket):
        client_logger.debug(f"CB: Socket open: socket: {socket}")

    def on_socket_close_callback(client, userdata, socket):
        client_logger.debug(f"CB: Socket close: socket: {socket}")

    def on_socket_register_write_callback(client, userdata, socket):
        client_logger.debug(f"CB: Socket register write: socket: {socket}")

    def on_socket_unregister_write_callback(client, userdata, socket):
        client_logger.debug(f"CB: Socket unregister write: socket: {socket}")

    mqtt_client = ClientSendsGracefulDisconnect(
        client_id="{}@{}[{}]".format(
            config.mqtt.CLIENT_ID_PREFIX,
            gethostname(),
            os.getpid(),
        ),
        clean_session=None,  # Required for MQTT5
        userdata=None,
        protocol=MQTTv5,
        transport=config.mqtt.TRANSPORT,
    )

    # mqtt_client.on_log = on_log_callback
    mqtt_client.on_connect = on_connect_callback
    # mqtt_client.on_publish = on_publish_callback
    mqtt_client.on_disconnect = on_disconnect_callback
    mqtt_client.on_socket_open = on_socket_open_callback
    mqtt_client.on_socket_close = on_socket_close_callback
    # mqtt_client.on_socket_register_write = on_socket_register_write_callback
    # mqtt_client.on_socket_unregister_write = on_socket_unregister_write_callback

    mqtt_client.enable_logger(client_logger)

    if config.mqtt.USERNAME is not None:
        logger.info(f"Connecting with username '{config.mqtt.USERNAME}'")
        mqtt_client.username_pw_set(
            username=config.mqtt.USERNAME,
            password=config.mqtt.PASSWORD
        )

    mqtt_client.will_set(topic=will_topic,
                         payload=MQTTStatusText.on_will.value,
                         qos=0,
                         retain=True,
                         properties=None
                         )

    if config.mqtt.TLS:
        mqtt_client.tls_set(ca_certs=config.mqtt.TLS_CA_CERTS,
                            certfile=config.mqtt.TLS_CERTFILE,
                            keyfile=config.mqtt.TLS_KEYFILE,
                            cert_reqs=config.mqtt.TLS_CERT_REQS,
                            tls_version=config.mqtt.TLS_VERSION,
                            ciphers=config.mqtt.TLS_CIPHERS)

    mqtt_client.connect(host=config.mqtt.BROKER_HOSTNAME,
                        port=config.mqtt.BROKER_PORT,
                        keepalive=config.mqtt.KEEPALIVE,
                        bind_address="",
                        bind_port=0,
                        clean_start=MQTT_CLEAN_START_FIRST_ONLY,
                        properties=None)

    mqtt_client.loop_start()

    last_update = time.time()
    update_period = 10  # in seconds
    counts = {}

    def create_pipe_reader_event_payload() -> Callable:

        def outbound_pipe_reader():

            nonlocal last_update, update_period, counts
            nonlocal outbound_pipe, mqtt_client

            item_json = outbound_pipe.recv()
            item_as_dict = json.loads(item_json)
            topic = f"{config.mqtt.TOPIC_ROOT}/{item_as_dict['class']}"
            mqtt_client.publish(
                topic=topic,
                payload=item_json,
                qos=0,
                # retain=True,  # Can cause client to always check if "current"
                retain=False,
                properties=None
            )

            now = time.time()
            try:
                counts[item_as_dict['class']] += 1
            except KeyError:
                counts[item_as_dict['class']] = 1
            if now - last_update > update_period:
                logger.getChild('Counts').debug(counts)
                counts = {}
                last_update = now

        return outbound_pipe_reader

    def create_pipe_reader_log_record() -> Callable:

        def outbound_pipe_reader():

            nonlocal last_update, update_period, counts
            nonlocal outbound_pipe, mqtt_client

            record: logging.LogRecord = outbound_pipe.recv()

            mqtt_client.publish(
                topic=f"{config.mqtt.TOPIC_ROOT}/log",
                payload=record.getMessage(),
                qos=0,
                retain=False,
                properties=None
            )

            record_as_json = pyde1_logging.log_record_to_json(record)
            mqtt_client.publish(
                topic=f"{config.mqtt.TOPIC_ROOT}/log/record",
                payload=record_as_json,
                qos=0,
                retain=False,
                properties=None
            )

        return outbound_pipe_reader

    if mode == OutboundMode.LogRecord:
        reader = create_pipe_reader_log_record()
    else:
        reader = create_pipe_reader_event_payload()

    loop.add_reader(outbound_pipe.fileno(), reader)

    loop.run_forever()



"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

from .event_manager import (
    SubscribedEvent, EventWithNotify, SequencerGate, send_to_outbound_pipes,
)
from .payloads import (
    EventPayload, EventNotificationName, SequencerGateName,
    SequencerGateNotification,
)



"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

This file is a part of the pyDE1 package.

These utilities appear to be in the public domain. Sources are identified below.
Copyright and of the original code is unknown and remains with the authors:

    https://stackoverflow.com/users/190597/unutbu
    https://gist.github.com/wonderbeyond

THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.
EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR
OTHER PARTIES PROVIDE THE PROGRAM “AS IS” WITHOUT WARRANTY OF ANY KIND,
EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.
SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL
NECESSARY SERVICING, REPAIR OR CORRECTION.

IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING WILL ANY
COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS THE PROGRAM
AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL,
SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE
OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF DATA
OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES
OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY
OF SUCH DAMAGES.

If the disclaimer of warranty and limitation of liability provided above
cannot be given local legal effect according to their terms, reviewing courts
shall apply local law that most closely approximates an absolute waiver
of all civil liability in connection with the Program, unless a warranty
or assumption of liability accompanies a copy of the Program in return for a fee.
"""

"""
Sources: 

https://stackoverflow.com/questions/31174295/getattr-and-setattr-on-nested-objects
https://gist.github.com/wonderbeyond/d293e7a2af1de4873f2d757edd580288

    "rgetattr and rsetattr are drop-in replacements for getattr and setattr, 
     which can also handle dotted attr strings."

"""

from functools import reduce


def rgetattr(obj, attr, *args):
    def _getattr(obj, attr):
        return getattr(obj, attr, *args)
    return reduce(_getattr, [obj] + attr.split('.'))


def rsetattr(obj, attr, val):
    pre, _, post = attr.rpartition('.')
    return setattr(rgetattr(obj, pre) if pre else obj, post, val)



"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

# Supervise:
#   Task: ResponseQueueProcessor
#   Task: RequestQueueProcessor

import asyncio
import logging
import multiprocessing.connection as mpc
import time
import traceback
from traceback import TracebackException

import pyDE1
from pyDE1.de1 import DE1
from pyDE1.dispatcher.implementation import (
    get_resource_to_dict, patch_resource_from_dict
)
from pyDE1.dispatcher.payloads import APIRequest, APIResponse, HTTPMethod
from pyDE1.dispatcher.resource import Resource
from pyDE1.exceptions import DE1NotConnectedError, DE1ValueError
from pyDE1.scale.processor import ScaleProcessor
from pyDE1.supervise import SupervisedTask

QUEUE_TOO_DEEP = 1  # If deeper than this, something is probably wrong, log

logger = pyDE1.getLogger('Inbound.Dispatcher')


# Interface from multiprocessing (sync) to asyncio.Queue() for async processing

def register_read_pipe_to_queue(
        pipe_to_read: mpc.Connection,
        queue_to_put: asyncio.Queue):
    asyncio.get_event_loop().add_reader(
        pipe_to_read.fileno(),
        _read_pipe_to_queue, pipe_to_read, queue_to_put)


def _read_pipe_to_queue(pipe_to_read: mpc.Connection,
                        queue_to_put: asyncio.Queue):
    data = pipe_to_read.recv()
    queue_to_put.put_nowait(data)
    if (qd := queue_to_put.qsize()) > QUEUE_TOO_DEEP:
        logger.error(
            f"Request queue exceeded QUEUE_TOO_DEEP, {qd} > {QUEUE_TOO_DEEP}")


def start_response_queue_processor(response_queue: asyncio.Queue,
                                   response_pipe: mpc.Connection):
    supervisor = SupervisedTask(
        _response_queue_processor,
        response_queue=response_queue,
        response_pipe=response_pipe)
    return supervisor


async def _response_queue_processor(response_queue: asyncio.Queue,
                                    response_pipe: mpc.Connection):

    while True:
        response = await response_queue.get()
        response_pipe.send(response)


# Process received APIRequests from the request queue,
# write to the outbound queue.

def start_request_queue_processor(request_queue: asyncio.Queue,
                                  response_queue: asyncio.Queue):
    supervisor = SupervisedTask(
        _request_queue_processor,
        request_queue=request_queue,
        response_queue=response_queue)
    return supervisor


async def _request_queue_processor(request_queue: asyncio.Queue,
                                   response_queue: asyncio.Queue):

    # Needed only to determine connectivity and readiness
    de1 = DE1()
    scale_processor = ScaleProcessor()

    def _check_connectivity(for_got: APIRequest,
                            check_de1 = True,
                            check_scale = True):
        if for_got.connectivity_required['DE1'] and check_de1:
            if not de1.is_connected:
                raise DE1NotConnectedError("DE1 not connected")
            elif not de1.is_ready:
                raise DE1NotConnectedError("DE1 not ready")
        if for_got.connectivity_required['Scale'] and check_scale:
            if scale_processor.scale is None:
                raise DE1NotConnectedError("No scale present")
            elif not scale_processor.scale.is_connected:
                raise DE1NotConnectedError("Scale not connected")
            elif not scale_processor.scale.is_ready:
                raise DE1NotConnectedError("Scale not ready")

    while True:
        got: APIRequest = await request_queue.get()
        logger.debug(f"{got.method.name} {got.resource.name} requires "
                     f"{got.connectivity_required}")
        resource_dict = {}
        exception = None
        tbe = None

        logger.debug(f"got: {got.method} {got.resource}")

        if got.method is HTTPMethod.GET:

            try:
                _check_connectivity(got)
                resource_dict = await get_resource_to_dict(got.resource)
            except Exception as e:
                exception = e
                tbe = TracebackException.from_exception(exception)
                if isinstance(exception, DE1NotConnectedError):
                    level = logging.INFO
                else:
                    level = logging.ERROR
                logger.log(level,
                           "Exception in processing "
                           f"{got.method} {got.resource} {repr(exception)}")
                logger.log(level,
                           ''.join(tbe.format()))

            response = APIResponse(original_timestamp=got.timestamp,
                                   timestamp=time.time(),
                                   payload=resource_dict,
                                   exception=exception,
                                   tbe=tbe)



        elif got.method is HTTPMethod.PATCH:

            if got.resource in (Resource.DE1_ID, Resource.SCALE_ID):
                if 'first_if_found' in got.payload:
                    # Only acceptable patch, if present
                    if len(got.payload.keys()) > 1:
                        raise DE1ValueError(
                            f"Use of 'first_if_found' with {got.resource} "
                            f"needs to be exclusive: {got.payload}")

            results_list = None
            check_de1 = True
            check_scale = True
            try:
                if (got.resource is Resource.DE1_ID
                    and len(got.payload.keys()) == 1
                    and ('id' in got.payload
                         or 'first_if_found' in got.payload)
                    and not de1.is_ready
                ):
                    logger.debug("DE1 gets a pass while disconnected")
                    check_de1 = False

                elif (got.resource is Resource.SCALE_ID
                        and len(got.payload.keys()) == 1
                        and ('id' in got.payload
                             or 'first_if_found' in got.payload)
                        and (scale_processor.scale is None
                             or not scale_processor.scale.is_ready)
                ):
                    logger.debug("Scale gets a pass while disconnected")
                    check_scale = False


                _check_connectivity(
                    got, check_de1=check_de1, check_scale=check_scale)
                results_list = await patch_resource_from_dict(got.resource,
                                                              got.payload)
            except Exception as e:
                exception = e
                tbe = TracebackException.from_exception(exception)
                logger.error(
                    f"Exception in processing {got.method} {got.resource}"
                    f" {repr(exception)}")
                logger.error(''.join(tbe.format()))

            response = APIResponse(original_timestamp=got.timestamp,
                                   timestamp=time.time(),
                                   payload=results_list,
                                   exception=exception,
                                   tbe=tbe)



        elif got.method is HTTPMethod.PUT:

            results_list = None
            try:
                if got.resource not in (Resource.DE1_PROFILE,
                                        Resource.DE1_PROFILE_ID,
                                        Resource.DE1_PROFILE_STORE,
                                        Resource.DE1_FIRMWARE):
                    raise NotImplementedError(
                        "Only profile and firmware PUT supported at this time")
                    # As there's no validation that a different PUT target
                    # is a complete replacement.

                # Profile store to database needs no device connectivity
                if got.resource == Resource.DE1_PROFILE_STORE:
                    check_de1 = False
                else:
                    check_de1 = True
                _check_connectivity(got, check_de1=check_de1)

                results_list = await patch_resource_from_dict(got.resource,
                                                              got.payload)
            except Exception as e:
                exception = e
                tbe = TracebackException.from_exception(exception)
                logger.error(
                    f"Exception in processing {got.method} {got.resource}"
                    f" {repr(exception)}")
                logger.error(''.join(tbe.format()))

            response = APIResponse(original_timestamp=got.timestamp,
                                   timestamp=time.time(),
                                   payload=results_list,
                                   exception=exception,
                                   tbe=tbe)

        else:

            response = APIResponse(original_timestamp=got.timestamp,
                                   timestamp=time.time(), payload={},
                                   exception=NotImplementedError(
                                       f"{got.method} is not supported"
                                   ))

        response_queue.put_nowait(response)
        if (qd := response_queue.qsize()) > QUEUE_TOO_DEEP:
            logger.error(
                "Response queue exceeded QUEUE_TOO_DEEP, "
                f"{qd} > {QUEUE_TOO_DEEP}")



"""
This class derived from that of Guido van Rossum available at
https://www.python.org/download/releases/2.2.3/descrintro/

As such, it appears from
https://docs.python.org/3/license.html#terms-and-conditions-for-accessing-or-otherwise-using-python
to be available under at least
https://docs.python.org/3/license.html#psf-license, retrieved from that URL
at the time this was originally added (2021-06-22)

1. This LICENSE AGREEMENT is between the Python Software Foundation ("PSF"), and
   the Individual or Organization ("Licensee") accessing and otherwise using Python
   3.9.5 software in source or binary form and its associated documentation.

2. Subject to the terms and conditions of this License Agreement, PSF hereby
   grants Licensee a nonexclusive, royalty-free, world-wide license to reproduce,
   analyze, test, perform and/or display publicly, prepare derivative works,
   distribute, and otherwise use Python 3.9.5 alone or in any derivative
   version, provided, however, that PSF's License Agreement and PSF's notice of
   copyright, i.e., "Copyright © 2001-2021 Python Software Foundation; All Rights
   Reserved" are retained in Python 3.9.5 alone or in any derivative version
   prepared by Licensee.

3. In the event Licensee prepares a derivative work that is based on or
   incorporates Python 3.9.5 or any part thereof, and wants to make the
   derivative work available to others as provided herein, then Licensee hereby
   agrees to include in any such work a brief summary of the changes made to Python
   3.9.5.

4. PSF is making Python 3.9.5 available to Licensee on an "AS IS" basis.
   PSF MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.  BY WAY OF
   EXAMPLE, BUT NOT LIMITATION, PSF MAKES NO AND DISCLAIMS ANY REPRESENTATION OR
   WARRANTY OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE
   USE OF PYTHON 3.9.5 WILL NOT INFRINGE ANY THIRD PARTY RIGHTS.

5. PSF SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF PYTHON 3.9.5
   FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL DAMAGES OR LOSS AS A RESULT OF
   MODIFYING, DISTRIBUTING, OR OTHERWISE USING PYTHON 3.9.5, OR ANY DERIVATIVE
   THEREOF, EVEN IF ADVISED OF THE POSSIBILITY THEREOF.

6. This License Agreement will automatically terminate upon a material breach of
   its terms and conditions.

7. Nothing in this License Agreement shall be deemed to create any relationship
   of agency, partnership, or joint venture between PSF and Licensee.  This License
   Agreement does not grant permission to use PSF trademarks or trade name in a
   trademark sense to endorse or promote products or services of Licensee, or any
   third party.

8. By copying, installing or otherwise using Python 3.9.5, Licensee agrees
   to be bound by the terms and conditions of this License Agreement.
"""


class Singleton:

    def __new__(cls, *args, **kwds):
        it = cls.__dict__.get("__it__")
        if it is not None:
            return it
        cls.__it__ = it = object.__new__(cls)
        it._singleton_init(*args, **kwds)
        return it

    def _singleton_init(self, *args, **kwds):
        pass

# NB: Original "init" changed to "_singleton_init" for clarity in subclasses

# To create a singleton class, you subclass from Singleton;
# each subclass will have a single instance,
# no matter how many times its constructor is called.
# To further initialize the subclass instance,
# subclasses should override 'init' instead of __init__
# - the __init__ method is called each time the constructor is called.
# For example:
#
# >>> class MySingleton(Singleton):
# ...     def init(self):
# ...         print "calling init"
# ...     def __init__(self):
# ...         print "calling __init__"
# ...
# >>> x = MySingleton()
# calling init
# calling __init__
# >>> assert x.__class__ is MySingleton
# >>> y = MySingleton()
# calling __init__
# >>> assert x is y
# >>>


"""
Copyright © 2021, 2022 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

For PATCH
  * Is writable
  * Any path is present in MAPPING[resource]
  * The type agrees with that in IsAt

For PUT (if implemented)
  * Everything for PATCH
  * All entries in MAPPING[resource] are present in the supplied data
"""

from typing import Union, get_args  # get_origin, get_type_hints

from pyDE1.dispatcher.mapping import MAPPING, IsAt
from pyDE1.dispatcher.resource import Resource
from pyDE1.exceptions import (
    DE1APIAttributeError, DE1APITypeError, DE1APIValueError
)


# t = Union[int, float]
# isinstance(2, t)
#   TypeError: Subscripted generics cannot be used with class and instance checks
# get_args(t)
#   (<class 'int'>, <class 'float'>)
# get_origin(t)
#   typing.Union
# isinstance(2, get_args(t))
#   True


def validate_patch_return_targets(resource: Resource,
                                  patch: Union[dict,
                                               bytes, bytearray]) -> dict:
    mapping = MAPPING[resource]

    # Valid: dict with dict
    #        IsAt with byte, bytearray (profile or firmware)
    if isinstance(mapping, dict) and isinstance(patch, dict):
        pass
    elif isinstance(mapping, IsAt) \
        and isinstance(patch, (bytes, bytearray)):
        # coerce into "standard form"
        patch = { None: patch }
    else:
        raise DE1APITypeError(
            "Validate: Mapping and patch inconsistent, "
            "dict with dict, IsAt with raw value "
            f"not {type(mapping)} with {type(patch)}"
        )

    results = {
        'DE1': False,
        'Scale': False
    }
    _validate_patch_inner(patch=patch,
                          mapping=mapping,
                          path='',
                          targets=results)
    return results


def _validate_patch_inner(patch: dict, mapping: dict, path: str, targets: dict):

    for key, new_value in patch.items():

        if path and len(path):
            this_path = f"{path}:{key}"
        else:
            this_path = key

        try:
            if key is not None:
                mapping_value = mapping[key]
            else:
                mapping_value = mapping
        except KeyError:
            raise DE1APIAttributeError(f"No mapping found for {this_path}")

        if isinstance(mapping_value, dict):
            _validate_patch_inner(
                patch=patch[key],
                mapping=mapping_value,
                path=this_path,
                targets=targets,
            )

        else:
            if not isinstance(mapping_value, IsAt):
                raise DE1APITypeError(
                    f"Expected an IsAt for {this_path}:, not {mapping_value}")

            if mapping_value.read_only:
                raise DE1APIAttributeError(f"Unable to write {this_path}:")

            # Check the value type
            # This will be a simple type, or something like Union, Optional
            # https://docs.python.org/3/library/typing.html#typing.get_args
            # get_args(a_simple_type) -> None

            # TODO: typing.ForwardRef -- For example, List["SomeClass"]
            #       NB: generic types such as list["SomeClass"]
            #       will not be implicitly transformed

            type_tuple = get_args(mapping_value.v_type)
            if len(type_tuple) == 0:
                type_tuple = (mapping_value.v_type,)
            if float in type_tuple and int not in type_tuple:
                # Accept an int for a float
                type_tuple = (*type_tuple, int,)

            if not isinstance(new_value, type_tuple):
                try:
                    type_str = mapping_value.v_type.__name__
                except AttributeError:
                    type_str = repr(mapping_value.v_type)
                raise DE1APITypeError(
                    f"Expected {type_str} value at {this_path}, "
                    f"not {new_value} {type(new_value)}"
                )

            if mapping_value.requires_connected_de1:
                targets['DE1'] = True
            if mapping_value.requires_connected_scale:
                targets['Scale'] = True

            # Not really "validate", but this is a good place to do it

            if (t := mapping_value.internal_type) is not None:
                try:
                    patch[key] = t(new_value)
                except ValueError as e:
                    raise DE1APIValueError(*e.args)



""""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

The BlueZ stack will keep a connection to the device open even if Python exits.
This wrapper "automatically" adds an atexit close of the connection, as well as
trying to remove that atexit command if the object's __del__ method gets called.

There is only a handful of these at any one time. Using the atexit handler
with unique calls for each should be sufficient, rather than managing a list.
"""

# Starting with bleak v0.18.0, the method signature changed
# for BleakScanner.find_device_by_filter()
# https://github.com/hbldh/bleak/issues/1028

BLEAK_AFTER_0_17 = False

from bleak import __version__ as bleak_version
bvs = bleak_version.split('.', 4)
bleak_major = 0
bleak_minor = 0
bleak_patch = 0
try:
    bleak_major = int(bvs[0])
    bleak_minor = int(bvs[1])
    bleak_patch = int(bvs[2])
except IndexError:
    pass
if int(bleak_major) > 0 or (int(bleak_major) == 0 and (bleak_minor) > 17):
    BLEAK_AFTER_0_17 = True




"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""
import asyncio
import logging
import sqlite3
import subprocess
from datetime import datetime
from multiprocessing import cpu_count
from pathlib import Path
from typing import Union, Optional

import pyDE1
from pyDE1.config import config
from pyDE1.exceptions import DE1DBError

logger = pyDE1.getLogger('Database.Manage')

CURRENT_USER_VERSION = 2
CURRENT_SCHEMA_RELPATH = 'schema/schema.002.sql'


def check_schema(loop: Optional[asyncio.AbstractEventLoop] = None):
    db_file_existed = Path(config.database.FILENAME).is_file()
    if not db_file_existed:
        logger.info(
            f"Database does not exist, will create {config.database.FILENAME}")
    try:
        with sqlite3.connect(config.database.FILENAME) as db:
            cur = db.execute('PRAGMA user_version')
            user_version = cur.fetchone()[0]
            if user_version == CURRENT_USER_VERSION:
                logger.debug(f"Confirmed user_version {user_version}")
            elif user_version:
                msg = (f"Database needs upgrade from {user_version} "
                       f"to {CURRENT_USER_VERSION}. Exiting.")
                logger.critical(msg)
                raise DE1DBError(msg)
            else:
                if db_file_existed:
                    bu_fname = '{}.{}.backup'.format(
                        config.database.FILENAME,
                        datetime.now().isoformat(timespec='seconds'),
                    )
                    backup_db(config.database.FILENAME, bu_fname,  loop)
                schema_path = Path(__file__).resolve().parent.joinpath(
                    CURRENT_SCHEMA_RELPATH).resolve()
                logger.info(
                    f"No user_version found. Installing schema from {schema_path}")
                schema = sql_commands_from_file(schema_path)
                logger.debug(f"Read {len(schema)} commands from {schema_path}")
                for sql in schema:
                    db.execute(sql)
                db.commit()
                logger.info("Initalization executed")
    except sqlite3.OperationalError as e:
        logger.critical(f"Database schema failed, aborting: {e}",
                        exc_info=e)
        raise DE1DBError(e)


def sql_commands_from_file(filename: Union[Path, str]) -> list:
    # This is "close" in that it should catch "common usage"
    # but might miss things, such as when literals contain the delimiters
    with open(filename, 'r') as sf:
        lines = sf.readlines()
    uncommented = []
    in_block_comment = False
    for line in lines:
        line = line.rstrip('\r\n')
        if not in_block_comment:
            idx_block_comment_start = line.find('/*')
            idx_eol_comment_start = line.find('--')
            if (    idx_block_comment_start == -1
                and idx_eol_comment_start == -1):
                if len(line.strip()):
                    uncommented.append(line)
            # From here on, one or both are present
            elif idx_block_comment_start == -1 \
                    or idx_eol_comment_start < idx_block_comment_start:
                # This is an EOL-style comment
                rest = line[0:idx_eol_comment_start]
                if len(rest.strip()):
                    uncommented.append(rest)
            elif idx_eol_comment_start == -1 \
                    or idx_block_comment_start < idx_eol_comment_start:
                # This is block-style comment start
                rest = line[0:idx_block_comment_start]
                if len(rest.strip()):
                    uncommented.append(rest)
                in_block_comment = True
            else:
                raise DE1DBError(
                    "Unexpected logic error in comment removal in SQL. "
                    f"idx_eol: {idx_eol_comment_start} "
                    f"idx_block: {idx_block_comment_start}")
        else:
            idx_block_comment_end = line.find('*/')
            if idx_block_comment_end != -1:
                uncommented.append(line[(idx_block_comment_end+2):])
                in_block_comment = False
            else:
                pass

    uncommented[-1] = (uncommented[-1].rstrip()).rstrip(';')

    # retval = ('\n'.join(uncommented)).split(';\n')
    return (' '.join(uncommented)).split('; ')


def backup_db(db_filename: str, backup_filename: str,
              loop: Optional[asyncio.AbstractEventLoop] = None):
    if Path(backup_filename).exists():
        msg = (f"Destination exists {backup_filename}, "
               f"aborting backup of {db_filename}")
        logger.critical(msg)
        raise DE1DBError(msg)

    logger.info(
        f"Backing up {db_filename} to {backup_filename}, "
        f"{config.database.BACKUP_TIMEOUT} s timeout."
    )

    try:
        cp = subprocess.run(
            ['sqlite3', db_filename, f".backup {backup_filename}"],
            capture_output=True,
            check=True,
            timeout=config.database.BACKUP_TIMEOUT,
        )
    except subprocess.CalledProcessError as e:
        logger.critical(
            f"Unable to backup {db_filename} to {backup_filename}, "
            f"{e}. Aborting\n"
            f"stderr: {e.stderr.decode('utf-8')}\n"
            f"stdout: {e.stdout.decode('utf-8')}\n"
        )
        raise DE1DBError(e)
    except subprocess.TimeoutExpired as e:
        logger.critical(
            f"Timeout during backup {db_filename} to {backup_filename}, "
            f"{e}. Aborting"
        )
        raise DE1DBError(e)

    logger.info(f"Backed up to {backup_filename}")

    if loop is not None:
        cpu_min = 2
        if cpu_count() >= cpu_min:
            logger.info(f"Scheduling compression of {backup_filename}")
            t_compress = loop.run_in_executor(
                None,
                subprocess.run,
                [config.database.BACKUP_COMPRESSION_EXECUTABLE,
                 backup_filename],
            )

            def tdc(fut: asyncio.Future):
                if fut.exception():
                    logger.warning("Compression task failed")
                else:
                    res: subprocess.CompletedProcess = fut.result()
                    if res.returncode:
                        logger.warning(
                            f"Compression execution error: {res.returncode}")

            t_compress.add_done_callback(tdc)
        else:
            logger.warning(
                f"NOT scheduling compression of {backup_filename} "
                f"with less than {cpu_min} CPU(s)")


if __name__ == '__main__':
    format_string = "%(asctime)s %(levelname)s %(name)s: %(message)s"

    logging.basicConfig(level=logging.DEBUG,
                        format=format_string,
                        )

    # pprint.pprint(
    #     sql_commands_from_file(
    #         '/home/ble-remote/devel/pyDE1/src/pyDE1/database/schema/schema.002.sql')
    # )

    config.load_from_yaml('/usr/local/etc/pyde1/pyde1.config')

    backup_db('/var/lib/pyde1/pyde1.sqlite3', '/home/ble-remote/bu.backup',
              asyncio.get_event_loop())


#!/usr/bin/env python3
"""
Copyright © 2022 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

Convert the de1app-written, Tcl-based profiles into JSON v2 format

Several fields that do not carry information, especially in the shot frames
are not present in this implementation.

Known limitations include:
  * No braces permitted in "text" fields (brace-quoted strings are OK)
"""

VERSION = {
    'app': '0.0.1',
    'format': '2.1',
}

import datetime
import os.path
import re
from pathlib import PurePath

import pyparsing
from pyparsing import (
    Suppress, Word,
    OneOrMore, ZeroOrMore,
    alphas, nums, alphanums, printables,
    common,
    dict_of,
    original_text_for,
)

from inspect import currentframe, getframeinfo
# Using PurePath here "normalizes" path
THIS_FILENAME = str(PurePath(getframeinfo(currentframe()).filename).name)
TIMESTAMP = datetime.datetime.now().astimezone().isoformat(timespec='seconds')

def pr_as_list(toks: pyparsing.results.ParseResults):
    return pyparsing.ParseResults.List(toks.as_list())

# This won't handle braces in the words

word_nb = Word(printables, exclude_chars='{}')


# Tcl dumps often leave representational errors
# A p16 data element is 1/16 scaled, 0.0625
def round2four(toks):
    val = toks[0] if isinstance(toks, pyparsing.ParseResults) else toks
    val = round(val, 4)
    ival = int(val)
    return ival if val == ival else val


rounded_number = common.number.copy().add_parse_action(round2four)

braced_string = (
      Suppress('{')
    + original_text_for(ZeroOrMore(word_nb))
    + Suppress('}')
)

valid_key = common.identifier

valid_simple_value = (
    rounded_number
    | word_nb
    | braced_string
)

shot_frame = (
      Suppress('{')
    + dict_of(valid_key, valid_simple_value)
    + Suppress('}')
)

shot_frame_as_dict = shot_frame.copy().set_parse_action(dict)

list_of_shot_frames = \
    OneOrMore(shot_frame_as_dict).set_parse_action(pr_as_list)

shot_frame_list =  (
      Suppress('{')
    + list_of_shot_frames
    + Suppress('}')
)

valid_profile_value = (
      common.number
    | word_nb
    | shot_frame_list
    | braced_string
)


profile_dict = dict_of(valid_key, valid_profile_value)


def show_test_result(result):
    if result[0]:
        print("\n------\n  OK  \n------\n")
    else:
        print("\n======\nFAILED\n======\n")


def braced_string_test():
    result = braced_string.run_tests(
        tests="""
        {}
        {one}
        {two words}    
        { one }
        { two words }
        """
    )
    show_test_result(result)

    result = braced_string.run_tests(
        failure_tests=True,
        tests="""
        { word {with internal brace} mess }
        # Embedded {one} word
        """
    )
    show_test_result(result)


def valid_key_test():
    result = valid_key.run_tests(
        tests="""
        word
        advanced_shot
        maximum_pressure_range_advanced
        """
    )
    show_test_result(result)

    result = valid_key.run_tests(
        failure_tests=True,
        tests="""
        two words
        {word}
        0
        3.4
        [foo]
        Downloaded from Visualizer}
        """
    )
    show_test_result(result)


def valid_simple_value_test():
    result = valid_simple_value.run_tests(
        tests="""
        word
        {}
        {one}
        {two words}
        {Extractamundo Tres!}  
        0
        -1
        1.2
        {0}
        {1.2}
        { 1.2 }
        {-3}
        { -4.5}
        """
    )
    show_test_result(result)

    result = valid_simple_value.run_tests(
        failure_tests=True,
        tests="""
        two words
        { word {with internal brace} mess }
        Embedded {one} word
        """
    )
    show_test_result(result)


def valid_shot_frame_test():
    test_data = [
        "{simple frame number 1}",
        "{simple frame number 2}",

        "{exit_if 1 flow 8.0 volume 100 max_flow_or_pressure_range 3.0 "
        "transition fast exit_flow_under 0 temperature 85.5 name {temp comp} "
        "pressure 8.0 sensor coffee pump pressure exit_type pressure_over "
        "exit_flow_over 6 exit_pressure_over 5.00 max_flow_or_pressure 0 "
        "exit_pressure_under 0 seconds 2.00}",

        "{exit_if 1 flow 8.0 volume 100 max_flow_or_pressure_range 3.0 "
        "transition fast exit_flow_under 0 temperature 80.5 weight 10.00 name "
        "preinfusion pressure 8.0 pump pressure sensor coffee exit_type "
        "pressure_over exit_flow_over 6 exit_pressure_over 5.00 "
        "max_flow_or_pressure 0 exit_pressure_under 0 seconds 20.00}",
    ]
    result = shot_frame.run_tests(
        full_dump=False,
        tests=test_data,
    )
    show_test_result(result)

    result = shot_frame.run_tests(
        failure_tests=True,
        full_dump=False,
        tests=[
            ' '.join(test_data[0:2]),
            ' '.join(test_data[2:4]),
        ],
    )
    show_test_result(result)


def valid_shot_frame_list_test():
    test_data = [
        "{{simple frame number 1}}",
        "{{simple frame number 1}{simple frame number 2}}",
        "{{simple frame number 1} {simple frame number 2}}",

        " {{exit_if 1 flow 8.0 volume 100 max_flow_or_pressure_range 3.0 "
        "transition fast exit_flow_under 0 temperature 85.5 name {temp comp} "
        "pressure 8.0 sensor coffee pump pressure exit_type pressure_over "
        "exit_flow_over 6 exit_pressure_over 5.00 max_flow_or_pressure 0 "
        "exit_pressure_under 0 seconds 2.00} {exit_if 1 flow 8.0 volume 100 "
        "max_flow_or_pressure_range 3.0 transition fast exit_flow_under 0 "
        "temperature 80.5 weight 10.00 name preinfusion pressure 8.0 pump "
        "pressure sensor coffee exit_type pressure_over exit_flow_over 6 "
        "exit_pressure_over 5.00 max_flow_or_pressure 0 exit_pressure_under "
        "0 seconds 20.00} {exit_if 1 flow 0 volume 100 "
        "max_flow_or_pressure_range 5.0 transition fast exit_flow_under 0 "
        "temperature 60.5 name {dynamic bloom} pressure 6.0 sensor coffee pump "
        "flow exit_type pressure_under exit_flow_over 6 max_flow_or_pressure 0 "
        "exit_pressure_over 11 exit_pressure_under 2.20 seconds 40.00} "
        "{exit_if 0 flow 6.0 volume 100 max_flow_or_pressure_range 5.0 "
        "transition fast exit_flow_under 0 temperature 60.5 name {6 mlps} "
        "pressure 6.0000000000000036 sensor coffee pump flow exit_type "
        "flow_under exit_flow_over 6 max_flow_or_pressure 2.0 "
        "exit_pressure_over 11 exit_pressure_under 0 seconds 60.00}}",
    ]
    test_data_fail = [
        "{}",
        "{{}}",

        "{exit_if 1 flow 8.0 volume 100 max_flow_or_pressure_range 3.0 "
        "transition fast exit_flow_under 0 temperature 85.5 name {temp comp} "
        "pressure 8.0 sensor coffee pump pressure exit_type pressure_over "
        "exit_flow_over 6 exit_pressure_over 5.00 max_flow_or_pressure 0 "
        "exit_pressure_under 0 seconds 2.00}",
    ]
    result = shot_frame_list.run_tests(
        # full_dump=False,
        tests=test_data,
    )
    show_test_result(result)

    result = shot_frame_list.run_tests(
        failure_tests=True,
        full_dump=False,
        tests=test_data_fail,
    )
    show_test_result(result)



def parsed_step_to_dict_v2(p_step: dict) -> dict:

    # p16 representation is 1/16 = 0.0625
    # so round to 4 digits to kill Tcl string errors

    # Pre-populate for readable order in JSON
    # will remove any None-valued from the step
    step_v2 = {
        'name':         p_step['name'],
        'pump':         p_step['pump'],
        'pressure':     None,
        'flow':         None,
        'limiter':      None,
        'transition':   p_step['transition'],
        'exit':         None,
        'weight':       None,
        'volume':       None,
        'seconds':      None,
        'temperature':  p_step['temperature'],
        'sensor':       p_step['sensor'],
    }

    if p_step['pump'] == 'pressure':
        step_v2['pressure'] = p_step['pressure']
    elif p_step['pump'] == 'flow':
        step_v2['flow'] = p_step['flow']
    else:
        raise ValueError(
            f"Unrecognized 'pump' type, '{p_step['pump']}")

    for move_on in ('seconds', 'volume', 'weight'):
        try:
            val = p_step[move_on]
            if val:
                step_v2[move_on] = val
        except KeyError:
            pass

    if p_step['exit_if']:
        exit_dict = dict()
        exit_type = p_step['exit_type']

        if exit_type == 'pressure_over':
            exit_dict['type'] = 'pressure'
            exit_dict['condition'] = 'over'
            exit_dict['value'] = p_step['exit_pressure_over']

        elif exit_type == 'pressure_under':
            exit_dict['type'] = 'pressure'
            exit_dict['condition'] = 'under'
            exit_dict['value'] = p_step['exit_pressure_under']

        elif exit_type == 'flow_over':
            exit_dict['type'] = 'flow'
            exit_dict['condition'] = 'over'
            exit_dict['value'] = p_step['exit_flow_over']

        elif exit_type == 'flow_under':
            exit_dict['type'] = 'flow'
            exit_dict['condition'] = 'under'
            exit_dict['value'] = p_step['exit_flow_under']

        else:
            raise ValueError(
                f"Unrecognized 'exit_type' {exit_type}")

        if exit_dict:
            exit_dict['value'] = exit_dict['value']
            step_v2['exit'] = exit_dict

    try:
        limiter_value = p_step['max_flow_or_pressure']
        if limiter_value:
            step_v2['limiter'] = {
                'value': limiter_value,
                'range': p_step['max_flow_or_pressure_range']
            }

    except KeyError:
        pass

    for k,v in step_v2.copy().items():
        if v is None:
            del step_v2[k]

    return step_v2


def parsed_dict_to_dict_v2(parsed: dict) -> dict:

    dict_v2 = {
        'version': VERSION['format'],
        'title': None,
        'notes': None,
        'author': None,
        'beverage_type': None,
        'steps': list(),
        'target_volume': None,
        'target_weight': None,
        'target_volume_count_start': None,
        'tank_temperature': None,
        'reference_file': None, # Eventually pick up from input file name
    }

    v2_key_map = {
        'author': 'author',
        'profile_title': 'title',
        'profile_notes': 'notes',
        'beverage_type': 'beverage_type',
        'tank_desired_water_temperature': 'tank_temperature',
        'final_desired_shot_weight_advanced': 'target_weight',
        'final_desired_shot_volume_advanced': 'target_volume',
        'final_desired_shot_volume_advanced_count_start':
            'target_volume_count_start',
        'settings_profile_type': 'legacy_profile_type',
        # 'type':	                            'type',	# 'advanced'
        'profile_language': 'lang',
        # 'profile_hide':	                    'hidden',	# 0/1
        # 'reference_file':	                    'reference_file',
        # 'changes_since_last_espresso':	    'changes_since_last_espresso',	# {}
        #                                       'version'   # 2,
    }

    for k,v in parsed.items():
        try:
            dict_v2[v2_key_map[k]] = v
        except KeyError:
            pass

    steps = dict_v2['steps']

    # A bit of history -- the types were named by John
    # based on the name the "screen" on which they were edited

    temperature_adjust_frame_time = 2  # seconds

    if parsed['settings_profile_type'] in ('settings_2c', 'settings_2c2'):
        dict_v2['type'] = 'advanced'
        for step in parsed['advanced_shot']:
            steps.append(parsed_step_to_dict_v2(step))

    elif parsed['settings_profile_type'] in ('settings_2a', 'settings_2b'):


        # espresso_temperature_0
        # need some parameter for frame 0 duration

        if parsed['espresso_temperature_1'] == parsed['espresso_temperature_0']:
            frame_0_duration = 0
        else:
            frame_0_duration = temperature_adjust_frame_time

        if parsed['preinfusion_time'] > frame_0_duration:
            frame_1_duration = round2four(
                parsed['preinfusion_time'] - frame_0_duration
            )
        else:
            frame_1_duration = 0

        steps = list()
        non_flow_frame_count = 0

        if frame_0_duration:
            steps.append(
                {
                    'name': 'temperature adj',
                    'pump': 'flow',
                    'flow': parsed['preinfusion_flow_rate'],
                    'transition': 'fast',
                    'exit': {
                        'type': 'pressure',
                        'condition': 'over',
                        'value': parsed['preinfusion_stop_pressure'],
                    },
                    'seconds': frame_0_duration,
                    'temperature': parsed['espresso_temperature_0'],
                    'sensor': 'coffee'
                },
            )
            non_flow_frame_count += 1

        if frame_1_duration:
            steps.append(
                {
                    'name': 'preinfuse',
                    'pump': 'flow',
                    'flow': parsed['preinfusion_flow_rate'],
                    'transition': 'fast',
                    'exit': {
                        'type': 'pressure',
                        'condition': 'over',
                        'value': parsed['preinfusion_stop_pressure'],
                    },
                    'seconds': frame_1_duration,
                    'temperature': parsed['espresso_temperature_1'],
                    'sensor': 'coffee'
                },
            )
            non_flow_frame_count += 1

            if parsed['settings_profile_type'] == 'settings_2a':

                dict_v2['type'] = 'pressure'
                if parsed['espresso_hold_time']:
                    steps.append(
                        {
                            'name': 'rise and hold',
                            'pump': 'pressure',
                            'pressure': parsed['espresso_pressure'],
                            'transition': 'smooth',
                            'limiter': {
                                'value': parsed['maximum_flow'],
                                'range': parsed['maximum_flow_range_default'],
                            },
                            'seconds': parsed['espresso_hold_time'],
                            'temperature': parsed['espresso_temperature_2'],
                            'sensor': 'coffee'
                        },
                    )

                if parsed['espresso_decline_time']:
                    steps.append(
                        {
                            'name': 'decline',
                            'pump': 'pressure',
                            'pressure': parsed['espresso_pressure'],
                            'transition': 'smooth',
                            'limiter': {
                                'value': parsed['maximum_flow'],
                                'range': parsed['maximum_flow_range_default'],
                            },
                            'seconds': parsed['espresso_decline_time'],
                            'temperature': parsed['espresso_temperature_3'],
                            'sensor': 'coffee'
                        },
                    )

            elif parsed['settings_profile_type'] == 'settings_2b':
                dict_v2['type'] = 'flow'

                if parsed['espresso_hold_time']:
                    steps.append(
                        {
                            'name': 'rise and hold',
                            'pump': 'flow',
                            'flow': parsed['flow_profile_hold'],
                            'transition': 'smooth',
                            'limiter': {
                                'value': parsed['maximum_pressure'],
                                'range': parsed[
                                    'maximum_pressure_range_default'],
                            },
                            'seconds': parsed['espresso_hold_time'],
                            'temperature': parsed['espresso_temperature_2'],
                            'sensor': 'coffee'
                        },
                    )

                if parsed['espresso_decline_time']:
                    steps.append(
                        {
                            'name': 'decline',
                            'pump': 'flow',
                            'flow': parsed['flow_profile_decline'],
                            'transition': 'smooth',
                            'limiter': {
                                'value': parsed['maximum_pressure'],
                                'range': parsed[
                                    'maximum_pressure_range_default'],
                            },
                            'seconds': parsed['espresso_decline_time'],
                            'temperature': parsed['espresso_temperature_3'],
                            'sensor': 'coffee'
                        },
                    )

            for step in steps:
                try:
                    if not step['limiter']['value']:
                        del step['limiter']
                except KeyError:
                    pass

            dict_v2['steps'] = steps
            dict_v2['target_volume_count_start'] = non_flow_frame_count
            dict_v2['target_volume'] = parsed['final_desired_shot_volume']
            dict_v2['target_weight'] = parsed['final_desired_shot_weight']

    dict_v2['creator'] = {
        'name': THIS_FILENAME,
        'version': VERSION['app'],
        'timestamp': TIMESTAMP,
    }


    return dict_v2


def dict_v2_set_author(dict_v2: dict[str], author: str):
    dict_v2['author'] = author


def dict_v2_set_reference_file(dict_v2: dict[str], filename: str):
    dict_v2['reference_file'] = filename


def dict_v2_get_title(dict_v2: dict[str]):
    return dict_v2['title']


def sanitize_filename(fname: str):
    return re.sub('[^\w._-]', '_', fname)


if __name__ == '__main__':

    import argparse
    import json
    import logging
    import sys

    from os.path import basename
    from pathlib import PurePath

    # import requests

    ap = argparse.ArgumentParser(
        description="Executable to open a Tcl profile file "
                    f"and write as JSON v{VERSION['format']}. "
                    "Input and output default to STDIN and STDOUT"
    )
    ap.add_argument('-a', '--author',
                    help='Replace author')
    input_group = ap.add_mutually_exclusive_group()
    input_group.add_argument('-i', '--input', help='Input file')
    input_group.add_argument('-v', '--visualizer', help='Visualizer URL')
    ap.add_argument('-o', '--output', help='Output file')
    ap.add_argument('-d', '--dir', help='Output directory')
    ap.add_argument('-f', '--force', action='store_true',
                    help='Overwrite if output exists')
    args = ap.parse_args()

    logger = logging.getLogger()

    formatter = logging.Formatter(
        "%(asctime)s %(levelname)s %(name)s: %(message)s")

    initial_handler = logging.StreamHandler(stream=sys.stderr)
    initial_handler.setFormatter(formatter)
    initial_handler.setLevel(logging.DEBUG)
    initial_handler.name = 'initial_logger'

    logger.addHandler(initial_handler)
    logger.setLevel(logging.DEBUG)

    ref_file = None

    if args.input is not None:
        ref_file = args.input
        with open(args.input, 'r') as fh:
            source_data = fh.read()

    elif args.visualizer is not None:
        ref_file = args.visualizer
        import requests
        logger.info(f"Getting {args.visualizer}")
        req = requests.get(url=args.visualizer)
        req.raise_for_status()
        source_data = req.text

    else:
        ref_file = None
        source_data = sys.stdin.read()


    pd_result = profile_dict.search_string(source_data)

    if len(pd_result) == 0:
        raise ValueError("No profile description found")
    elif len(pd_result) > 1:
        logging.error(
            "Multiple profile descriptions found. Ignoring all except first")
    pdd = dict(pd_result[0])
    dv2 = parsed_dict_to_dict_v2(pdd)

    if args.author is not None:
        dict_v2_set_author(dv2, args.author)

    if ref_file is not None:
        dict_v2_set_reference_file(dv2, ref_file)

    # Output file names:
    #
    # I V O D
    # -------
    #           stdout              0
    #       D   dir/<title>.json
    #     O     output              2
    #     O D   dir/output          1
    #   V       stdout              0
    #   V   D   dir/<title>.json
    #   V O     output              2
    #   V 0 D   dir/output          1
    # I         stdout              0
    # I     D   dir/input.json      3
    # I   O     output              2
    # I   O D   dir/output          1
    # I V x x   Not possible        -

    # Rule 0 -- to stdout
    if args.output is None and args.dir is None:
        fname = None
        print(json.dumps(dv2, indent=2), file=sys.stdout)

    # Rule 1 -- to dir/output
    elif args.output is not None:
        fname = PurePath(args.dir, args.output)

    # Rule 2 -- to output
    elif args.output is not None:
        fname = args.output

    # Rule 3 -- from input name to dir
    elif args.input is not None:
        fname = PurePath(args.dir, basename(args.input)).with_suffix('.json')

    else:
        profile_name = dict_v2_get_title(dv2)
        fname = sanitize_filename(profile_name) + '.json'


    if fname is not None:
        if os.path.exists(fname):
            if args.force:
                logger.warning(f"Overwriting {fname}")
            else:
                raise FileExistsError(fname)
        with open(fname, 'w') as fh:
            print(json.dumps(dv2, indent=2), file=fh)
        logger.info(f"Output written to {fname}")



import enum
import json
import time
import uuid
from typing import Optional

import pyDE1
from pyDE1.utils import prep_for_json

logger = pyDE1.getLogger('EventManager.Payloads')

class EventPayload:
    """
    A canonical event payload, passed to Event.publish()
    and received by each of the subscribers as their argument.

    Additional attributes can be added by subclasses to pass data

    arrival_time    is mandatory, represents when the "trigger" occurred
    create_time     if None, will use time.time()
    _sender         will be filled out by the Event.publish() method
    _event_time     will be filled out by the Event.publish() method
    """
    _internal_only = False

    def __init__(self,
                 arrival_time: float,
                 create_time: Optional[float] = None,
                 ):
        self._version = None
        self._sender = None
        self.arrival_time = arrival_time
        if create_time is None:
            create_time = time.time()
        self.create_time = create_time
        self._event_time = None

    @property
    def version(self):
        return self._version

    @property
    def sender(self):
        return self._sender

    @property
    def event_time(self):
        return self._event_time

    # Keep signature consistent with PackedAttr.as_wire_bytes()
    def as_json(self):
        """
        Convert to JSON for external consumers.
        Consumer is responsible for "wrapping" this payload for delivery.

        The sender is converted to the name of the sender's class.
        __name__ should return the base name, though would return,
        for example "AtomaxSkaleII" and not "Scale"
        Times are in time.time() format, 1623096954.5960422

        Only _name, _version, _sender and _event_time
        are accepted from "private" attributes.
        They are translated to 'name', 'version', 'sender' and 'event_time'
        """
        # IntEnum gets JSON-ified as an int
        work = {k: prep_for_json(v) for k, v in self.__dict__.items()
                if not k.startswith('_')}
        for key in ('version', 'event_time'):
            try:
                work[key] = self.__dict__['_' + key]
            except KeyError:
                work[key] = None
        work['sender'] = type(self._sender).__name__
        work['class'] = type(self).__name__

        return json.dumps(work)


class EventNotificationName (enum.Enum):
    """
    As "subclassing an enumeration is allowed only
        if the enumeration does not define any members."
    all need to be defined here
    """
    pass


class EventNotificationAction (enum.Enum):
    SET = 'set'
    CLEAR = 'clear'


class EventNotification (EventPayload):

    def __init__(self, arrival_time: Optional[float],
                 create_time: Optional[float] = None,
                 sender = None,
                 name: EventNotificationName = None,
                 action: EventNotificationAction = None
                 ):
        if not isinstance(name, EventNotificationName):
            raise TypeError(
                "EventNotification needs a valid EventNotificationName, not "
                f"'{name}'")
        if not isinstance(action, EventNotificationAction):
            raise TypeError(
                "EventNotification needs a valid EventNotificationAction, not "
                f"'{action}'")
        if arrival_time is None:
            arrival_time = time.time()
        super(EventNotification, self).__init__(
            arrival_time=arrival_time,
            create_time=create_time,
        )
        self._version = "1.0.0"
        self._sender = sender
        self.name = name.value
        self.action = action.value


class SequencerGateName (EventNotificationName):
    GATE_SEQUENCE_START = "sequence_start"
    GATE_FLOW_BEGIN = "sequence_flow_begin"
    GATE_EXPECT_DROPS = "sequence_expect_drops"
    GATE_EXIT_PREINFUSE = "sequence_exit_preinfuse"
    GATE_FLOW_END = "sequence_flow_end"
    GATE_FLOW_STATE_EXIT = "sequence_flow_state_exit"
    GATE_LAST_DROPS = "sequence_last_drops"
    GATE_SEQUENCE_COMPLETE = "sequence_complete"


class SequencerGateNotification (EventNotification):

    sequence_id = uuid.uuid4()

    @classmethod
    def new_sequence(cls):
        cls.sequence_id = str(uuid.uuid4())
        return cls.sequence_id

    def __init__(self, arrival_time: Optional[float],
                 create_time: Optional[float] = None,
                 sender = None,
                 name: EventNotificationName = None,
                 action: EventNotificationAction = None
                 ):
        super(SequencerGateNotification, self).__init__(
            arrival_time=arrival_time,
            create_time=create_time,
            sender=sender,
            name=name,
            action=action
        )
        self._version = "1.1.0"
        self.sequence_id = self.__class__.sequence_id
        # active_state gets set by the SequencerGate which has a reference
        # to the FlowSequencerImpl, which implements FlowSequencer
        # Otherwise there's a messy circular import
        self.active_state = None    # Gets set


"""
Copyright © 2021, 2022 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

See `manual_setup()` for some still-needed setup on process start
"""

import multiprocessing
import multiprocessing.connection as mpc

import pyDE1.config


def run_controller(master_config: pyDE1.config.Config,
                   log_queue: multiprocessing.Queue,
                   inbound_pipe: mpc.Connection,
                   outbound_pipe: mpc.Connection,
                   database_queue: multiprocessing.Queue):

    pyDE1.config.config = master_config
    from pyDE1.config import config

    import asyncio
    import time

    import pyDE1.pyde1_logging as pyde1_logging
    import pyDE1.shutdown_manager as sm

    from pyDE1.de1 import DE1
    from pyDE1.de1.c_api import API_MachineStates
    from pyDE1.dispatcher.dispatcher import (
        register_read_pipe_to_queue,
        start_request_queue_processor, start_response_queue_processor
    )
    from pyDE1.event_manager import SubscribedEvent
    from pyDE1.flow_sequencer import FlowSequencer
    from pyDE1.scale.processor import ScaleProcessor

    pyde1_logging.setup_queue_logging(config.logging, log_queue)
    pyde1_logging.config_logger_levels(config.logging)

    logger = pyDE1.getLogger('Controller')

    loop = asyncio.get_event_loop()
    loop.set_debug(True)

    def on_shutdown_underway_cleanup():
        logger.info("Watching for shutdown event")
        sm.shutdown_underway.wait()

        async def _the_rest():
            t0 = time.time()
            de1 = DE1()

            logger.debug(
                f"DE1 current state: {de1.current_state.__repr__()}")
            if de1.current_state not in (
                    API_MachineStates.Sleep,
                    API_MachineStates.GoingToSleep,
                    API_MachineStates.NoRequest,
            ):
                if de1.is_connected:
                    logger.info("Sleep DE1")
                    await de1.sleep()
                else:
                    logger.warning("Unable to sleep DE1, not connected")
            logger.info(f"Disconnecting devices")
            # for device in _disconnect_set:
            #     await device.disconnect()
            for device in [DE1(), ScaleProcessor().scale]:
                if device is not None:
                    await device.disconnect()
            t1 = time.time()
            logger.info(f"Controller shutdown took {t1 - t0:0.3f} sec")
            logger.info("Setting cleanup_complete")
            sm.cleanup_complete.set()

        asyncio.run_coroutine_threadsafe(_the_rest(), loop)

    on_shutdown_wait_task = loop.run_in_executor(
        None, on_shutdown_underway_cleanup)

    sm.attach_signal_handler_to_loop(sm.shutdown, loop)

    loop.set_exception_handler(sm.exception_handler)

    request_queue = asyncio.Queue()
    response_queue = asyncio.Queue()

    register_read_pipe_to_queue(
        pipe_to_read=inbound_pipe,
        queue_to_put=request_queue,
    )

    # In dispatcher, this process "does the work"
    start_request_queue_processor(request_queue=request_queue,
                                  response_queue=response_queue)

    # In dispatcher, moves response from queue to pipe
    start_response_queue_processor(
        response_queue=response_queue,
        response_pipe=inbound_pipe
    )

    # Sets up the destination for events to be sent to outbound (MQTT) API
    SubscribedEvent.outbound_pipe = outbound_pipe
    SubscribedEvent.database_queue = database_queue

    # TODO: This may no longer be robust, make a classmethod to set/get?
    FlowSequencer.database_queue = database_queue

    loop.run_forever()



"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

import asyncio
import logging
import time
from typing import Union

import pyDE1
from pyDE1 import de1
from pyDE1.de1.ble import CUUID
from pyDE1.de1.c_api import (
    Versions, RequestedState, SetTime, ReadFromMMR, WriteToMMR,
    Temperatures, ShotSettings, ShotSample, StateInfo, HeaderWrite, FrameWrite,
    WaterLevels, FWMapRequest, MMR0x80LowAddr, API_MachineStates, API_Substates,
    Calibration, CalTargets, CalCommand
)
from pyDE1.de1.events import StateUpdate, ShotSampleUpdate, WaterLevelUpdate
from pyDE1.exceptions import DE1ErrorStateReported, MMRAddressOffsetError
from pyDE1.utils import data_as_readable_or_hex, data_as_hex

# Logging is set to DEBUG by default. This effectively disables them
# with independent control. (The evaluation of the f-string is still done)
for cuuid in CUUID:
    pyDE1.getLogger(
        f"DE1.{cuuid.__str__()}.Notify").setLevel(logging.INFO)

def default_handler_map(de1: de1):
    return {
        CUUID.Versions: create_Versions_callback(de1),
        CUUID.RequestedState: create_RequestedState_callback(de1),
        CUUID.SetTime: create_SetTime_callback(de1),
        CUUID.ShotDirectory: create_ShotDirectory_callback(de1),
        CUUID.ReadFromMMR: create_ReadFromMMR_callback(de1),
        CUUID.WriteToMMR: create_WriteToMMR_callback(de1),
        CUUID.ShotMapRequest: create_ShotMapRequest_callback(de1),
        CUUID.DeleteShotRange: create_DeleteShotRange_callback(de1),
        CUUID.FWMapRequest: create_FWMapRequest_callback(de1),
        CUUID.Temperatures: create_Temperatures_callback(de1),
        CUUID.ShotSettings: create_ShotSettings_callback(de1),
        CUUID.Deprecated: create_Deprecated_callback(de1),
        CUUID.ShotSample: create_ShotSample_callback(de1),
        CUUID.StateInfo: create_StateInfo_callback(de1),
        CUUID.HeaderWrite: create_HeaderWrite_callback(de1),
        CUUID.FrameWrite: create_FrameWrite_callback(de1),
        CUUID.WaterLevels: create_WaterLevels_callback(de1),
        CUUID.Calibration: create_Calibration_callback(de1),
    }

"""
bleak takes either standard functions or coroutines

        if inspect.iscoroutinefunction(callback):

            def bleak_callback(s, d):
                asyncio.ensure_future(callback(s, d))

        else:
            bleak_callback = callback


From the Python docs
    Important See also the create_task() function 
    which is the preferred way for creating new Tasks.
    

So it looks like a coroutine will be thrown on the loop

GO WITH THAT ASSUMPTION
"""

def create_Versions_callback(de1: de1):

    async def Versions_callback(sender: int, data: Union[bytes, bytearray]):
        nonlocal de1
        arrival_time = time.time()
        obj = Versions().from_wire_bytes(data, arrival_time)
        de1._cuuid_dict[CUUID.Versions].mark_updated(obj, arrival_time)
        logger = pyDE1.getLogger(f"DE1.{CUUID.Versions.__str__()}.Notify")
        logger.debug(obj.log_string())
    return Versions_callback


def create_RequestedState_callback(de1: de1):

    async def RequestedState_callback(sender: int, data: Union[bytes, bytearray]):
        nonlocal de1
        arrival_time = time.time()
        obj = RequestedState().from_wire_bytes(data, arrival_time)
        de1._cuuid_dict[CUUID.RequestedState].mark_updated(obj, arrival_time)
        logger = pyDE1.getLogger(f"DE1.{CUUID.RequestedState.__str__()}.Notify")
        logger.debug(obj.log_string())
    return RequestedState_callback


def create_SetTime_callback(de1: de1):

    async def SetTime_callback(sender: int, data: Union[bytes, bytearray]):
        nonlocal de1
        arrival_time = time.time()
        obj = SetTime().from_wire_bytes(data, arrival_time)
        de1._cuuid_dict[CUUID.SetTime].mark_updated(obj, arrival_time)
        logger = pyDE1.getLogger(f"DE1.{CUUID.SetTime.__str__()}.Notify")
        logger.debug(obj.log_string())
    return SetTime_callback


def create_ShotDirectory_callback(de1: de1):

    async def ShotDirectory_callback(sender: int, data: Union[bytes, bytearray]):
        nonlocal de1
        arrival_time = time.time()
        # obj = ShotDirectory().from_wire_bytes(data, arrival_time)
        # logger.debug(obj.log_string())
        de1._cuuid_dict[CUUID.ShotDirectory].mark_updated(data, arrival_time)
        logger = pyDE1.getLogger(f"DE1.{CUUID.ShotDirectory.__str__()}.Notify")
        logger.debug(f"{data_as_readable_or_hex(data)} ({len(data)})")
    return ShotDirectory_callback


def create_ReadFromMMR_callback(de1: de1):

    async def ReadFromMMR_callback(sender: int, data: Union[bytes, bytearray]):
        nonlocal de1
        arrival_time = time.time()

        # A read of CUUID.ReadFromMMR returns the request ReadFromMMR
        # not the "usual" data from the MMR that is "notified" here
        # For this case, Len means number of words - 1 for the request
        # It needs to be special-cased here
        # and probably doesn't need the read-on-write behavior

        # It's not obvious how to know this is reading back
        # from CUUID.ReadFromMMR as the request has the target address

        obj = ReadFromMMR().from_wire_bytes(data, arrival_time,
                                            from_response=True)
        # Logging of the full response is intentionally
        # early as MMR may contain multiple registers
        logger = pyDE1.getLogger(f"DE1.{CUUID.ReadFromMMR.__str__()}.Notify")
        logger.debug(obj.log_string())
        if obj.addr_high != 0x80:
            # Can't write these to de1._mmr_dict as it assumes 0x80
            logger.error(
                "Unhandled MMR response from "
                f"0x{obj.addr_high:02x} {obj.addr_low:04x}"
            )

        elif obj.is_within_debug_log:
            # If completely within the debug-log region, take as a whole
            # Not being actively handled right now
            notify_state = de1._mmr_dict[obj.addr_low]
            notify_state.data_raw = obj.Data
            notify_state.mark_updated(obj, arrival_time)
            ds = data_as_readable_or_hex(notify_state.data_decoded,
                                         replace_rnt=True)
            logger.debug(f"Debug log: {ds}")

        elif (obj.Len % 4 != 0) or (obj.addr_low % 4 != 0):
            raise MMRAddressOffsetError(
                "Only MMR stride of 4 bytes is implemented at this time."
                f"len(mmr_bytes) bytes at 0x{obj.addr_low:04x} passed."
            )

        else:
            # Split into four-byte segments and process
            start = 0
            while (start + 4) <= obj.Len:
                this_addr = obj.addr_low + start
                notify_state = de1._mmr_dict[this_addr]
                notify_state.data_raw = obj.Data[start:(start + 4)]
                # This ensures that the data is decoded before Event.set()
                notify_state.mark_updated(notify_state.data_raw, arrival_time)
                mmr = MMR0x80LowAddr.for_logging(this_addr, return_as_hex=True)
                logger.getChild(mmr).info(notify_state.data_decoded)
                start += 4
        de1._cuuid_dict[CUUID.ReadFromMMR].mark_updated(obj, arrival_time)
    return ReadFromMMR_callback


def create_WriteToMMR_callback(de1: de1):

    async def WriteToMMR_callback(sender: int, data: Union[bytes, bytearray]):
        nonlocal de1
        arrival_time = time.time()

        # TODO: This needs to manage from_response as well
        #       Can they both be handled in MMRData?

        obj = WriteToMMR().from_wire_bytes(data, arrival_time)
        de1._cuuid_dict[CUUID.WriteToMMR].mark_updated(obj, arrival_time)
        logger = pyDE1.getLogger(f"DE1.{CUUID.WriteToMMR.__str__()}.Notify")
        logger.debug(obj.log_string())
    return WriteToMMR_callback


def create_ShotMapRequest_callback(de1:de1):

    async def ShotMapRequest_callback(sender: int, data: Union[bytes, bytearray]):
        nonlocal de1
        arrival_time = time.time()
        # obj = ShotMapRequest().from_wire_bytes(data, arrival_time)
        # logger.debug(obj.log_string())
        de1._cuuid_dict[CUUID.ShotMapRequest].mark_updated(data, arrival_time)
        logger = pyDE1.getLogger(f"DE1.{CUUID.ShotMapRequest.__str__()}.Notify")
        logger.debug(f"{data_as_readable_or_hex(data)} ({len(data)})")
    return ShotMapRequest_callback


def create_DeleteShotRange_callback(de1: de1):

    async def DeleteShotRange_callback(sender: int, data: Union[bytes, bytearray]):
        nonlocal de1
        arrival_time = time.time()
        # obj = DeleteShotRange().from_wire_bytes(data, arrival_time)
        # logger.debug(obj.log_string())
        de1._cuuid_dict[CUUID.DeleteShotRange].mark_updated(data, arrival_time)
        logger = pyDE1.getLogger(f"DE1.{CUUID.DeleteShotRange.__str__()}.Notify")
        logger.debug(f"{data_as_readable_or_hex(data)} ({len(data)})")
    return DeleteShotRange_callback


def create_FWMapRequest_callback(de1: de1):

    async def FWMapRequest_callback(sender: int, data: Union[bytes, bytearray]):
        nonlocal de1
        arrival_time = time.time()
        obj = FWMapRequest().from_wire_bytes(data, arrival_time)
        de1._cuuid_dict[CUUID.FWMapRequest].mark_updated(obj, arrival_time)
        logger = pyDE1.getLogger(f"DE1.{CUUID.FWMapRequest.__str__()}.Notify")
        logger.debug(obj.log_string())
    return FWMapRequest_callback


def create_Temperatures_callback(de1: de1):

    async def Temperatures_callback(sender: int, data: Union[bytes, bytearray]):
        nonlocal de1
        arrival_time = time.time()
        obj = Temperatures().from_wire_bytes(data, arrival_time)
        de1._cuuid_dict[CUUID.Temperatures].mark_updated(obj, arrival_time)
        logger = pyDE1.getLogger(f"DE1.{CUUID.Temperatures.__str__()}.Notify")
        logger.debug(obj.log_string())
    return Temperatures_callback


def create_ShotSettings_callback(de1: de1):

    async def ShotSettings_callback(sender: int, data: Union[bytes, bytearray]):
        nonlocal de1
        arrival_time = time.time()
        obj = ShotSettings().from_wire_bytes(data, arrival_time)
        de1._cuuid_dict[CUUID.ShotSettings].mark_updated(obj, arrival_time)
        logger = pyDE1.getLogger(f"DE1.{CUUID.ShotSettings.__str__()}.Notify")
        logger.debug(obj.log_string())
    return ShotSettings_callback


def create_Deprecated_callback(de1: de1):

    async def Deprecated_callback(sender: int, data: Union[bytes, bytearray]):
        nonlocal de1
        arrival_time = time.time()
        # obj = Deprecated().from_wire_bytes(data, arrival_time)
        # logger.debug(obj.log_string())
        de1._cuuid_dict[CUUID.Deprecated].mark_updated(data, arrival_time)
        logger = pyDE1.getLogger(f"DE1.{CUUID.Deprecated.__str__()}.Notify")
        logger.error(f"{data_as_readable_or_hex(data)} ({len(data)})")
    return Deprecated_callback


def create_ShotSample_callback(de1: de1):

    async def ShotSample_callback(sender: int, data: Union[bytes, bytearray]):
        nonlocal de1
        arrival_time = time.time()
        obj = ShotSample().from_wire_bytes(data, arrival_time)
        de1._cuuid_dict[CUUID.ShotSample].mark_updated(obj, arrival_time)
        await de1._event_shot_sample.publish(
            ShotSampleUpdate(
                arrival_time=arrival_time,
                sample_time=obj.SampleTime,
                group_pressure=obj.GroupPressure,
                group_flow=obj.GroupFlow,
                mix_temp=obj.MixTemp,
                head_temp=obj.HeadTemp,
                set_mix_temp=obj.SetMixTemp,
                set_head_temp=obj.SetHeadTemp,
                set_group_pressure=obj.SetGroupPressure,
                set_group_flow=obj.SetGroupFlow,
                frame_number=obj.FrameNumber,
                steam_temp=obj.SteamTemp,
            ))
        logger = pyDE1.getLogger(f"DE1.{CUUID.ShotSample.__str__()}.Notify")
        logger.debug(obj.log_string())
    return ShotSample_callback


def create_StateInfo_callback(de1: de1):
    """
    NB: This can block itself.

    The working assumption is that a coroutine is thrown at the loop
    and that the "fairness" of the lock will keep things in order
    """

    previous_state = API_MachineStates.NoRequest
    previous_substate = API_Substates.NoState
    previous_lock = asyncio.Lock()

    async def StateInfo_callback(sender: int, data: Union[bytes, bytearray]):
        nonlocal de1, previous_state, previous_substate, previous_lock

        arrival_time = time.time()
        obj = StateInfo().from_wire_bytes(data, arrival_time)
        de1._cuuid_dict[CUUID.StateInfo].mark_updated(obj, arrival_time)

        if obj.State == API_MachineStates.FatalError or obj.SubState.is_error:
            details = f"DE1 reported error condition: {obj.log_string()}"
            logger = pyDE1.getLogger(f"DE1.{CUUID.StateInfo.__str__()}.Notify")
            logger.error(details)
            raise DE1ErrorStateReported(details)

        # Keep the DE1 up to date as quickly s possible
        de1._current_state = obj.State
        de1._current_substate = obj.SubState

        async with previous_lock:
            # Put the message onto the loop to be able to drop the lock faster
            await de1._event_state_update.publish(
                StateUpdate(
                    arrival_time=arrival_time,
                    state=obj.State,
                    substate=obj.SubState,
                    previous_state=previous_state,
                    previous_substate=previous_substate,
                ))
            previous_state = obj.State
            previous_substate = obj.SubState

        logger = pyDE1.getLogger(f"DE1.{CUUID.StateInfo.__str__()}.Notify")
        logger.debug(obj.log_string())
    return StateInfo_callback


def create_HeaderWrite_callback(de1: de1):

    async def HeaderWrite_callback(sender: int, data: Union[bytes, bytearray]):
        nonlocal de1
        arrival_time = time.time()
        obj = HeaderWrite().from_wire_bytes(data, arrival_time)
        de1._cuuid_dict[CUUID.HeaderWrite].mark_updated(obj, arrival_time)
        logger = pyDE1.getLogger(f"DE1.{CUUID.HeaderWrite.__str__()}.Notify")
        logger.debug(obj.log_string())
    return HeaderWrite_callback


def create_FrameWrite_callback(de1: de1):

    async def FrameWrite_callback(sender: int, data: Union[bytes, bytearray]):
        nonlocal de1
        arrival_time = time.time()
        obj = FrameWrite().from_wire_bytes(data, arrival_time)
        de1._cuuid_dict[CUUID.FrameWrite].mark_updated(obj, arrival_time)
        logger = pyDE1.getLogger(f"DE1.{CUUID.FrameWrite.__str__()}.Notify")
        logger.debug(obj.log_string())
    return FrameWrite_callback


def create_WaterLevels_callback(de1: de1):

    async def WaterLevels_callback(sender: int, data: Union[bytes, bytearray]):
        nonlocal de1
        arrival_time = time.time()
        obj = WaterLevels().from_wire_bytes(data, arrival_time)
        de1._cuuid_dict[CUUID.WaterLevels].mark_updated(obj, arrival_time)
        await de1._event_water_levels.publish(
            WaterLevelUpdate(
                arrival_time=arrival_time,
                level=obj.Level,
                start_fill_level=obj.StartFillLevel,
            ))
        logger = pyDE1.getLogger(f"DE1.{CUUID.WaterLevels.__str__()}.Notify")
        logger.debug(obj.log_string())
    return WaterLevels_callback


def create_Calibration_callback(de1:de1):

    async def Calibration_callback(sender: int, data: Union[bytes, bytearray]):
        nonlocal de1
        arrival_time = time.time()
        obj = Calibration().from_wire_bytes(data, arrival_time)
        # TODO: This notifies multiple values from the same CUUID
        de1._cuuid_dict[CUUID.Calibration].mark_updated(obj, arrival_time)
        logger = pyDE1.getLogger(f"DE1.{CUUID.Calibration.__str__()}.Notify")
        cal_set = None
        setter = None

        if obj.CalCommand in (CalCommand.Read, CalCommand.Write):
            cal_set = '_cal_local'
        else:
            cal_set = '_cal_factory'

        if obj.CalTarget == CalTargets.CalFlow:
            setter = 'record_flow'
        elif obj.CalTarget == CalTargets.CalPressure:
            setter = 'record_pressure'
        elif obj.CalTarget == CalTargets.CalTemp:
            setter = 'record_temperature'
        elif obj.CalTarget == CalTargets.CalError:
            setter = None
            logger.error("Calibration returned error result, not setting")

        if setter is not None:
            cal = de1.__getattribute__(cal_set)
            method = cal.__getattribute__(setter)
            method(de1_value=obj.DE1ReportedVal, measured=obj.MeasuredVal)
        logger.info(obj.log_string())
    return Calibration_callback





"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

Deciding no how to handle "global" config is a bit more complicated
for multiprocessing than it is for a single process (with or without threads).

Since each process loads Python, to get the same config values for all processes
they all need to either read the same config file, or create copies of the
"original" for themselves. Using something like a multiprocessing.Manager()
seems excessive, as it is unlikely that the values will change during a run.
Leaving that open for the future makes some sense.

That means that each process will need to be either passed the name of the
config file to be read, or the results of that (and any changes prior to
creation of the child processes). Passing the config instance seems to be
no more complex than passing the file reference.

The config data should be easy to access in each of the modules.
Several options for "singleton" behavior:

1) Use a bare module

Advantages:

    Simple access

        import pyDE1.config as config
        do_something_with(config.section.VALUE)

    Only one instance "guaranteed"

Disadvantages:

    No top-level properties

        Though there may be a way to define them, the @property decorator
        will create a property object, but it doesn't seem useful.

        This is one of the main reasons that the Singleton object is used
        for the primary components of pyDE1 (the other is no dangling instances)


2) Module that defines an instance, refer to the instance

Advantages:

    Simple access

        from pyDE1.config import config
        do_something_with(config.section.VALUE)

Disadvantages:

    It is possible that second instance, disconnected from the "real" one,
    could be generated with config = Config()

3) Subclass Singleton

Advantages:

    Only one instance "guaranteed"

Disadvantages:

    More complex access

        from pyDE1.config import Config
        pointer_to_config = Config()    # At the module or function level
        do_something_with(config.section.VALUE)



"""

import logging
from typing import Optional

import pyDE1
from pyDE1.config_load import ConfigYAML, ConfigLoadable
from pyDE1.pyde1_logging import (
    ConfigLogging, ConfigLoggingFormatters, ConfigLoggingHandlers
)

DEFAULT_CONFIG_FILE = '/usr/local/etc/pyde1/pyde1.conf'

logger = pyDE1.getLogger('Config')


class Config (ConfigYAML):

    DEFAULT_CONFIG_FILE = '/usr/local/etc/pyde1/pyde1.conf'

    def __init__(self):
        super(Config, self).__init__()
        self.bluetooth = _Bluetooth()
        self.database = _Database()
        self.de1 = _DE1()
        self.http = _HTTP(self)    # Calculating timeout needs bluetooth
        self.logging = _Logging()
        self.mqtt = _MQTT()


# This craziness is so pyCharm autocompletes
# Otherwise typing.SimpleNamespace() would be sufficient


class _MQTT (ConfigLoadable):
    def __init__(self):
        self.TOPIC_ROOT = 'pyDE1'
        self.CLIENT_ID_PREFIX = 'pyde1'
        self.BROKER_HOSTNAME = '::1'
        self.BROKER_PORT = 1883
        self.TRANSPORT = 'tcp'
        self.KEEPALIVE = 60
        self.USERNAME = None
        self.PASSWORD = None
        self.DEBUG = False
        self.TLS = False    # Set True, or rest of TLS is ignored
        # See paho Client.tls_set() for details
        self.TLS_CA_CERTS = None
        self.TLS_CERTFILE = None
        self.TLS_KEYFILE = None
        self.TLS_CERT_REQS = None
        self.TLS_VERSION = None
        self.TLS_CIPHERS = None


class _HTTP (ConfigLoadable):
    def __init__(self, parent):
        self.SERVER_HOST = ''
        self.SERVER_PORT = 1234
        self.SERVER_ROOT = '/'
        # adaptive_allonge.json is 7632 bytes
        self.PATCH_SIZE_LIMIT = 16384
        # Seconds, before abandoning the request
        self.ASYNC_TIMEOUT = 1.0
        # Seconds, 20*2 frames + head + tail at ~100 ms each
        self.PROFILE_TIMEOUT = 4.5
        self.FIRMWARE_TIMEOUT = 15  # Seconds for upload and start (~260 kbps)
        self._response_timeout = None

        # If true, don't output nodes that have no value (write-only)
        # or are empty dicts
        # Otherwise math.nan fills in for the missing value
        # As not compliant with RFC 7159, some parsers may fail with NaN
        # although it is permitted by ECMAScript and JavaScript
        # A False setting is intended to be a development/exploration tool
        # This feature be considered as deprecated
        self.PRUNE_EMPTY_NODES = True

        self._parent = parent   # Path to get to bluetooth

    @property
    def RESPONSE_TIMEOUT(self):
        # See pyDE1/dispatcher/implementation.py
        # Right now, single timeout, bounded by scan/connect
        # This is in addition to the timeout in the implementation
        if self._response_timeout:
            retval = self._response_timeout
        else:
            retval = (max((self._parent.bluetooth.SCAN_TIME
                           + self._parent.bluetooth.CONNECT_TIMEOUT
                           + self.ASYNC_TIMEOUT),
                          (self.PROFILE_TIMEOUT
                           + self.ASYNC_TIMEOUT
                           + self._parent.de1.CUUID_LOCK_WAIT_TIMEOUT))
                      + 0.100)
        return retval

    @RESPONSE_TIMEOUT.setter
    def RESPONSE_TIMEOUT(self, value):
        self._response_timeout = value


class _Logging (ConfigLogging):
    def __init__(self):
        super(_Logging, self).__init__()


class _LoggingFormatters (ConfigLoggingFormatters):
    def __init__(self):
        super(_LoggingFormatters, self).__init__()


class _LoggingHandlers (ConfigLoggingHandlers):
    def __init__(self):
        super(_LoggingHandlers, self).__init__()


class _Bluetooth (ConfigLoadable):
    def __init__(self):
        self.SCAN_TIME = 5  # Seconds
        self.CONNECT_TIMEOUT = 10  # Seconds
        self.DISCONNECT_TIMEOUT = 5  # Seconds
        self.SCAN_CACHE_EXPIRY = 300  # Seconds, probably too long
        self.RECONNECT_RETRY_COUNT = 10 # Before using RECONNECT_GAP
        self.RECONNECT_GAP = 10 # Seconds
        # Files that hold the Bluetooth ID of connected devices
        # for potential cleanup by supervisor scripts
        self.ID_FILE_DIRECTORY = '/var/lib/pyde1/'
        self.ID_FILE_SUFFIX = '.btid'


class _Database (ConfigLoadable):
    def __init__(self):
        self.FILENAME = '/var/lib/pyde1/pyde1.sqlite3'
        self.BACKUP_TIMEOUT = 60  # seconds
        self.BACKUP_COMPRESSION_EXECUTABLE = 'xz'


class _DE1 (ConfigLoadable):
    def __init__(self):
        self.LINE_FREQUENCY = 60
        self.MAX_WAIT_FOR_READY_EVENTS = 4.0 # Seconds (28 at 0.1 each)
        self.CUUID_LOCK_WAIT_TIMEOUT = 2 # Seconds
        self.SEQUENCE_WATCHDOG_TIMEOUT = 270 # seconds
        self.DEFAULT_AUTO_OFF_TIME = None   # Minutes
        self.STOP_AT_WEIGHT_ADJUST = -0.07  # Secs, larger increases weight
        self.bump_resist = _BumpResist()
        self.API_STOP_IGNORES_CHECKS = False  # Request Idle in all cases
        self.PATCH_ON_CONNECT = None  # If defined as a dict, PATCH /de1


class _BumpResist (ConfigLoadable):
    def __init__(self):
        # For stop/skip based on weight, try to ignore "bumps"
        # If the estimated weight flow is over the threshold
        # use multiplier * DE1 estimated flow
        self.FLOW_THRESHOLD = 10.0  # g/s
        self.FLOW_MULTIPLIER = 1.1
        self.SUB_MEDIAN_WEIGHT = True  # when excessive flow
        self.USE_MEDIAN_WEIGHT_ALWAYS = False
        self.USE_MEDIAN_FLOW_ALWAYS = False


config = Config()



"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""


class DE1Error (RuntimeError):
    pass


class DE1ValueError (DE1Error, ValueError):
    pass


class DE1TypeError (DE1Error, TypeError):
    pass


class DE1RuntimeError (DE1Error, RuntimeError):
    pass


class DE1AttributeError(DE1Error, AttributeError):
    pass


class DE1NoAddressError (DE1Error):
    pass


class DE1NotConnectedError (DE1Error):
    pass


class DE1IsConnectedError (DE1Error):
    """
    Such as trying to wipe state while connected
    """
    pass


class DE1NoHandlerError(DE1Error):
    pass


class DE1ErrorStateReported(DE1Error):
    pass


class DE1OperationInProgressError(DE1Error):
    pass


class DE1APIError (DE1Error):
    pass


class DE1APITypeError (DE1APIError, TypeError):
    pass


class DE1APIValueError (DE1APIError, ValueError):
    def __init__(self, *args, **kwargs):
        super(DE1APIValueError, self).__init__(*args, **kwargs)


class DE1APIUnsupportedFeatureError (DE1APIValueError):
    pass


# Pickle of custom exceptions discussed at
# https://stackoverflow.com/questions/16244923/how-to-make-a-custom-exception-class-with-multiple-init-args-pickleable

class DE1APIUnsupportedStateTransitionError (DE1APIValueError):
    def __init__(self, target_mode, current_state, current_substate,
                 *args, **kwargs):
        # Formatting the message for the super call still fails un-pickle
        try:
            current_state = current_state.name
        except AttributeError:
            pass
        try:
            current_substate = current_substate.name
        except AttributeError:
            pass
        self.target_mode = target_mode
        self.current_state = current_state
        self.current_substate = current_substate
        super(DE1APIUnsupportedStateTransitionError, self).__init__(
            "I'm afraid I can't do that Dave. Can't move to "
            f"{target_mode} from {current_state}, {current_substate}",
            *args, **kwargs)

    def __reduce__(self):
        return (DE1APIUnsupportedStateTransitionError, (self.target_mode,
                                                        self.current_state,
                                                        self.current_substate))


class DE1APIAttributeError (DE1APIError, AttributeError):
    pass


class DE1APINotManagedHereException (DE1APIAttributeError):
    """
    Used to indicate to the FlowSequencer that a different parameter
    is used to manage the behavior. Currently used for steam time.
    """
    pass


class DE1APIKeyError(DE1APIError, KeyError):
    pass


class DE1APITooManyFramesError (DE1APIValueError):
    pass


class MMRTypeError (DE1APITypeError):
    pass


class MMRValueError (DE1APIValueError):
    pass


class MMRDataTooLongError (MMRValueError):
    pass


class MMRAddressError (MMRValueError):
    pass


class MMRAddressRangeError (MMRAddressError):
    pass


class MMRAddressOffsetError (MMRAddressError):
    pass


class DE1DBError (RuntimeError):
    pass

class DE1IncompleteSequenceRecordError (DE1DBError):
    pass

class DE1DBNoMatchingRecord (DE1DBError):
    pass

class DE1DBConsistencyError (DE1DBError):
    pass


"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

Message definitions for database/write_notifications queue
to break cyclical imports
"""
from typing import NamedTuple


class RecorderControl (NamedTuple):
    recording: bool
    sequence_id: str



"""
Copyright © 2021, 2022 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

Main process for receiving update packets and logging to the database
"""

# Only import the minimal here, as it potentially ends up in all processes.

import multiprocessing

import pyDE1.config


def run_database_logger(master_config: pyDE1.config.Config,
                        log_queue: multiprocessing.Queue,
                        notification_queue: multiprocessing.Queue):

    pyDE1.config.config = master_config
    from pyDE1.config import config

    import asyncio
    import threading

    import pyDE1.pyde1_logging as pyde1_logging
    import pyDE1.shutdown_manager as sm

    # Failing if only here (and not at the module level)
    from pyDE1.database.write_notifications import record_data

    from pyDE1.supervise import SupervisedTask

    logger = pyDE1.getLogger('Database.Logger')

    pyde1_logging.setup_queue_logging(config.logging, log_queue)
    pyde1_logging.config_logger_levels(config.logging)

    loop = asyncio.get_event_loop()
    loop.set_debug(True)

    def on_shutdown_underway_cleanup():
        logger.info("Watching for shutdown event")
        sm.shutdown_underway.wait()
        logger.info("Shutdown cleanup start")
        logger.info(f"Threads: {threading.enumerate()}")

        async def _the_rest():
            t = 1.1
            logger.info(f"Trying a {t} second sleep")
            await asyncio.sleep(t)
            logger.info(f"Threads: {threading.enumerate()}")
            logger.info("Setting cleanup_complete")
            sm.cleanup_complete.set()

        asyncio.run_coroutine_threadsafe(_the_rest(), loop)

    on_shutdown_wait_task = loop.run_in_executor(
        None, on_shutdown_underway_cleanup)

    sm.attach_signal_handler_to_loop(sm.shutdown, loop)

    loop.set_exception_handler(sm.exception_handler)

    async def heartbeat():
        hlog = pyDE1.getLogger('Heartbeat.Database.Logger')
        while True:
            await asyncio.sleep(10)
            hlog.debug("===== BRAP =====")

    SupervisedTask(heartbeat)

    SupervisedTask(record_data, notification_queue)

    loop.run_forever()